@@ -28,7 +28,9 @@
 #include <sys/stat.h>
 #include <sys/types.h>
 #ifndef _WIN32
+#include <sys/wait.h>
 #include <pwd.h>
+#include <grp.h>
 #endif
 
 #include <libxml/parser.h>
@@ -966,13 +968,14 @@ static int fetch_config(struct openconnect_info *vpninfo)
 
 static int run_csd_script(struct openconnect_info *vpninfo, char *buf, int buflen)
 {
-#ifdef _WIN32
+#if defined(_WIN32) || defined(__native_client__)
 	vpn_progress(vpninfo, PRG_ERR,
-		     _("Error: Running the 'Cisco Secure Desktop' trojan on Windows is not yet implemented.\n"));
+		     _("Error: Running the 'Cisco Secure Desktop' trojan on this platform is not yet implemented.\n"));
 	return -EPERM;
 #else
 	char fname[64];
 	int fd, ret;
+	pid_t child;
 
 	if (!vpninfo->csd_wrapper && !buflen) {
 		vpn_progress(vpninfo, PRG_ERR,
@@ -1042,80 +1045,115 @@ static int run_csd_script(struct openconnect_info *vpninfo, char *buf, int bufle
 		close(fd);
 	}
 
-	if (!fork()) {
-		char scertbuf[MD5_SIZE * 2 + 1];
-		char ccertbuf[MD5_SIZE * 2 + 1];
-		char *csd_argv[32];
-		int i = 0;
+	child = fork();
+	if (child == -1) {
+		goto out;
+	} else if (child > 0) {
+		/* in parent: must reap child process */
+		int status;
+		waitpid(child, &status, 0);
+	} else {
+		/* in child: run CSD script as daemon */
+		if (fork()) {
+			/* child must use _exit(2) */
+			_exit(0);
+		} else {
+			/* in grandchild: will be reaped by init */
+			char scertbuf[MD5_SIZE * 2 + 1];
+			char ccertbuf[MD5_SIZE * 2 + 1];
+			char *csd_argv[32];
+			int i = 0;
 
-		if (vpninfo->uid_csd_given && vpninfo->uid_csd != getuid()) {
-			struct passwd *pw;
+			setsid();
 
-			if (setuid(vpninfo->uid_csd)) {
-				fprintf(stderr, _("Failed to set uid %ld\n"),
-					(long)vpninfo->uid_csd);
-				exit(1);
+			if (vpninfo->uid_csd_given && vpninfo->uid_csd != getuid()) {
+				struct passwd *pw;
+				int e;
+
+				if (setgid(vpninfo->gid_csd)) {
+					e = errno;
+					fprintf(stderr, _("Failed to set gid %ld: %s\n"),
+						(long)vpninfo->uid_csd, strerror(e));
+					exit(1);
+				}
+
+				if (setgroups(1, &vpninfo->gid_csd)) {
+					e = errno;
+					fprintf(stderr, _("Failed to set groups to %ld: %s\n"),
+						(long)vpninfo->uid_csd, strerror(e));
+					exit(1);
+				}
+
+				if (setuid(vpninfo->uid_csd)) {
+					e = errno;
+					fprintf(stderr, _("Failed to set uid %ld: %s\n"),
+						(long)vpninfo->uid_csd, strerror(e));
+					exit(1);
+				}
+
+				if (!(pw = getpwuid(vpninfo->uid_csd))) {
+					e = errno;
+					fprintf(stderr, _("Invalid user uid=%ld: %s\n"),
+						(long)vpninfo->uid_csd, strerror(e));
+					exit(1);
+				}
+				setenv("HOME", pw->pw_dir, 1);
+				if (chdir(pw->pw_dir)) {
+					e = errno;
+					fprintf(stderr, _("Failed to change to CSD home directory '%s': %s\n"),
+						pw->pw_dir, strerror(e));
+					exit(1);
+				}
 			}
-			if (!(pw = getpwuid(vpninfo->uid_csd))) {
-				fprintf(stderr, _("Invalid user uid=%ld\n"),
-					(long)vpninfo->uid_csd);
-				exit(1);
-			}
-			setenv("HOME", pw->pw_dir, 1);
-			if (chdir(pw->pw_dir)) {
-				fprintf(stderr, _("Failed to change to CSD home directory '%s': %s\n"),
-					pw->pw_dir, strerror(errno));
-				exit(1);
+			if (getuid() == 0 && !vpninfo->csd_wrapper) {
+				fprintf(stderr, _("Warning: you are running insecure "
+						  "CSD code with root privileges\n"
+						  "\t Use command line option \"--csd-user\"\n"));
 			}
+			/* Spurious stdout output from the CSD trojan will break both
+			   the NM tool and the various cookieonly modes. */
+			dup2(2, 1);
+			if (vpninfo->csd_wrapper)
+				csd_argv[i++] = openconnect_utf8_to_legacy(vpninfo,
+									   vpninfo->csd_wrapper);
+			csd_argv[i++] = fname;
+			csd_argv[i++] = (char *)"-ticket";
+			if (asprintf(&csd_argv[i++], "\"%s\"", vpninfo->csd_ticket) == -1)
+				goto out;
+			csd_argv[i++] = (char *)"-stub";
+			csd_argv[i++] = (char *)"\"0\"";
+			csd_argv[i++] = (char *)"-group";
+			if (asprintf(&csd_argv[i++], "\"%s\"", vpninfo->authgroup?:"") == -1)
+				goto out;
+
+			openconnect_local_cert_md5(vpninfo, ccertbuf);
+			scertbuf[0] = 0;
+			get_cert_md5_fingerprint(vpninfo, vpninfo->peer_cert, scertbuf);
+			csd_argv[i++] = (char *)"-certhash";
+			if (asprintf(&csd_argv[i++], "\"%s:%s\"", scertbuf, ccertbuf) == -1)
+				goto out;
+
+			csd_argv[i++] = (char *)"-url";
+			if (asprintf(&csd_argv[i++], "\"https://%s%s\"", vpninfo->hostname, vpninfo->csd_starturl) == -1)
+				goto out;
+
+			csd_argv[i++] = (char *)"-langselen";
+			csd_argv[i++] = NULL;
+
+			if (setenv("CSD_TOKEN", vpninfo->csd_token, 1))
+				goto out;
+			if (setenv("CSD_HOSTNAME", vpninfo->hostname, 1))
+				goto out;
+
+			apply_script_env(vpninfo->csd_env);
+
+			execv(csd_argv[0], csd_argv);
+
+		out:
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Failed to exec CSD script %s\n"), csd_argv[0]);
+			exit(1);
 		}
-		if (getuid() == 0 && !vpninfo->csd_wrapper) {
-			fprintf(stderr, _("Warning: you are running insecure "
-					  "CSD code with root privileges\n"
-					  "\t Use command line option \"--csd-user\"\n"));
-		}
-		/* Spurious stdout output from the CSD trojan will break both
-		   the NM tool and the various cookieonly modes. */
-		dup2(2, 1);
-		if (vpninfo->csd_wrapper)
-			csd_argv[i++] = openconnect_utf8_to_legacy(vpninfo,
-								   vpninfo->csd_wrapper);
-		csd_argv[i++] = fname;
-		csd_argv[i++] = (char *)"-ticket";
-		if (asprintf(&csd_argv[i++], "\"%s\"", vpninfo->csd_ticket) == -1)
-			goto out;
-		csd_argv[i++] = (char *)"-stub";
-		csd_argv[i++] = (char *)"\"0\"";
-		csd_argv[i++] = (char *)"-group";
-		if (asprintf(&csd_argv[i++], "\"%s\"", vpninfo->authgroup?:"") == -1)
-			goto out;
-
-		openconnect_local_cert_md5(vpninfo, ccertbuf);
-		scertbuf[0] = 0;
-		get_cert_md5_fingerprint(vpninfo, vpninfo->peer_cert, scertbuf);
-		csd_argv[i++] = (char *)"-certhash";
-		if (asprintf(&csd_argv[i++], "\"%s:%s\"", scertbuf, ccertbuf) == -1)
-			goto out;
-
-		csd_argv[i++] = (char *)"-url";
-		if (asprintf(&csd_argv[i++], "\"https://%s%s\"", vpninfo->hostname, vpninfo->csd_starturl) == -1)
-			goto out;
-
-		csd_argv[i++] = (char *)"-langselen";
-		csd_argv[i++] = NULL;
-
-		if (setenv("CSD_TOKEN", vpninfo->csd_token, 1))
-			goto out;
-		if (setenv("CSD_HOSTNAME", vpninfo->hostname, 1))
-			goto out;
-
-		apply_script_env(vpninfo->csd_env);
-
-		execv(csd_argv[0], csd_argv);
-
-out:
-		vpn_progress(vpninfo, PRG_ERR,
-			     _("Failed to exec CSD script %s\n"), csd_argv[0]);
-		exit(1);
 	}
 
 	free(vpninfo->csd_stuburl);
@@ -1129,7 +1167,7 @@ out:
 
 	http_add_cookie(vpninfo, "sdesktop", vpninfo->csd_token, 1);
 	return 0;
-#endif /* !_WIN32 */
+#endif /* !_WIN32 && !__native_client__ */
 }
 
 
