@@ -129,18 +129,6 @@ static void calculate_mtu(struct openconnect_info *vpninfo, int *base_mtu, int *
 		*mtu = 1280;
 }
 
-/* For OpenSSL the configure script detects DTLS 1.2 support.
- * For GnuTLS just check for v3.2.0+ */
-#if defined(DTLS_GNUTLS) && GNUTLS_VERSION_NUMBER >= 0x030200
-#define HAVE_DTLS12 1
-#endif
-
-#ifdef HAVE_DTLS12
-# define DEFAULT_CIPHER_LIST "OC-DTLS1_2-AES256-GCM:OC-DTLS1_2-AES128-GCM:AES256-SHA:AES128-SHA:DES-CBC3-SHA:DES-CBC-SHA"
-#else
-# define DEFAULT_CIPHER_LIST "AES256-SHA:AES128-SHA:DES-CBC3-SHA:DES-CBC-SHA"
-#endif
-
 static void append_compr_types(struct oc_text_buf *buf, const char *proto, int avail)
 {
 	if (avail) {
@@ -227,6 +215,7 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 			       vpninfo->disable_ipv6 ? "IPv4" : "IPv6,IPv4");
 	if (!vpninfo->disable_ipv6)
 		buf_append(reqbuf, "X-CSTP-Full-IPv6-Capability: true\r\n");
+#ifdef HAVE_DTLS
 	if (vpninfo->dtls_state != DTLS_DISABLED) {
 		buf_append(reqbuf, "X-DTLS-Master-Secret: ");
 		for (i = 0; i < sizeof(vpninfo->dtls_secret); i++) {
@@ -239,11 +228,16 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 			buf_free(reqbuf);
 			return -EINVAL;
 		}
-		buf_append(reqbuf, "\r\nX-DTLS-CipherSuite: %s\r\n",
-			   vpninfo->dtls_ciphers ? : DEFAULT_CIPHER_LIST);
+		buf_append(reqbuf, "\r\nX-DTLS-CipherSuite: ");
+		if (vpninfo->dtls_ciphers)
+			buf_append(reqbuf, "%s", vpninfo->dtls_ciphers);
+		else
+			append_dtls_ciphers(vpninfo, reqbuf);
+		buf_append(reqbuf, "\r\n");
 
 		append_compr_types(reqbuf, "DTLS", vpninfo->req_compr & ~COMPR_DEFLATE);
 	}
+#endif
 	buf_append(reqbuf, "\r\n");
 
 	if (buf_error(reqbuf)) {
@@ -502,7 +496,8 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 	}
 	vpninfo->ip_info.mtu = mtu;
 
-	if (!vpninfo->ip_info.addr && !vpninfo->ip_info.addr6) {
+	if (!vpninfo->ip_info.addr && !vpninfo->ip_info.addr6 &&
+	    !vpninfo->ip_info.netmask6) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("No IP address received. Aborting\n"));
 		return -EINVAL;
@@ -831,7 +826,7 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 	   and add POLLOUT. As it is, though, it'll just chew CPU time in that
 	   fairly unlikely situation, until the write backlog clears. */
 	while (1) {
-		int len = vpninfo->deflate_pkt_size ? : vpninfo->ip_info.mtu;
+		int len = MAX(16384, vpninfo->deflate_pkt_size ? : vpninfo->ip_info.mtu);
 		int payload_len;
 
 		if (!vpninfo->cstp_pkt) {
