@@ -25,6 +25,10 @@
 #include <errno.h>
 #include <stdlib.h>
 #include <stdio.h>
+#ifndef _WIN32
+#include <netinet/in.h>
+#include <sys/socket.h>
+#endif
 
 #include "openconnect-internal.h"
 
@@ -333,11 +337,18 @@ static int dtls_try_handshake(struct openconnect_info *vpninfo)
 	int ret = SSL_do_handshake(vpninfo->dtls_ssl);
 
 	if (ret == 1) {
+		const char *c;
 		vpninfo->dtls_state = DTLS_CONNECTED;
 		vpn_progress(vpninfo, PRG_INFO,
 			     _("Established DTLS connection (using OpenSSL). Ciphersuite %s.\n"),
 			     vpninfo->dtls_cipher);
 
+		c = openconnect_get_dtls_compression(vpninfo);
+		if (c) {
+			vpn_progress(vpninfo, PRG_INFO,
+				     _("DTLS connection compression using %s.\n"), c);
+		}
+
 		vpninfo->dtls_times.last_rekey = vpninfo->dtls_times.last_rx = 
 			vpninfo->dtls_times.last_tx = time(NULL);
 
@@ -354,7 +365,12 @@ static int dtls_try_handshake(struct openconnect_info *vpninfo)
 		   (this is one of the areas in which Cisco's DTLS differs
 		   from the RFC4347 spec), and DPD should help us notice
 		   if *nothing* is getting through. */
-#if OPENSSL_VERSION_NUMBER >= 0x1000005fL
+#if OPENSSL_VERSION_NUMBER >= 0x10100000L
+		/* OpenSSL 1.1.0 or above. Do nothing. The SSLeay() function
+		   got renamed, and it's a pointless check in this case
+		   anyway because there's *no* chance that we linked against
+		   1.1.0 and are running against something older than 1.0.0e. */
+#elif OPENSSL_VERSION_NUMBER >= 0x1000005fL
 		/* OpenSSL 1.0.0e or above doesn't resend anyway; do nothing.
 		   However, if we were *built* against 1.0.0e or newer, but at
 		   runtime we find that we are being run against an older
@@ -434,31 +450,112 @@ void dtls_shutdown(struct openconnect_info *vpninfo)
 	SSL_CTX_free(vpninfo->dtls_ctx);
 }
 
+void append_dtls_ciphers(struct openconnect_info *vpninfo, struct oc_text_buf *buf)
+{
+#ifdef HAVE_DTLS12
+	buf_append(buf, "OC-DTLS1_2-AES256-GCM:OC-DTLS1_2-AES128-GCM:AES256-SHA:AES128-SHA:DES-CBC3-SHA:DES-CBC-SHA");
+#else
+	buf_append(buf, "AES256-SHA:AES128-SHA:DES-CBC3-SHA:DES-CBC-SHA");
+#endif
+}
+
 #elif defined(DTLS_GNUTLS)
 #include <gnutls/dtls.h>
 #include "gnutls.h"
 
+#if GNUTLS_VERSION_NUMBER < 0x030200
+# define GNUTLS_DTLS1_2 202
+#endif
+#if GNUTLS_VERSION_NUMBER < 0x030400
+# define GNUTLS_CIPHER_CHACHA20_POLY1305 23
+#endif
+
+static void detect_mtu(struct openconnect_info *vpninfo);
+
 struct {
 	const char *name;
 	gnutls_protocol_t version;
 	gnutls_cipher_algorithm_t cipher;
+	gnutls_kx_algorithm_t kx;
 	gnutls_mac_algorithm_t mac;
 	const char *prio;
+	const char *min_gnutls_version;
 } gnutls_dtls_ciphers[] = {
-	{ "AES128-SHA", GNUTLS_DTLS0_9, GNUTLS_CIPHER_AES_128_CBC, GNUTLS_MAC_SHA1,
-	  "NONE:+VERS-DTLS0.9:+COMP-NULL:+AES-128-CBC:+SHA1:+RSA:%COMPAT" },
-	{ "AES256-SHA", GNUTLS_DTLS0_9, GNUTLS_CIPHER_AES_256_CBC, GNUTLS_MAC_SHA1,
-	  "NONE:+VERS-DTLS0.9:+COMP-NULL:+AES-256-CBC:+SHA1:+RSA:%COMPAT" },
-	{ "DES-CBC3-SHA", GNUTLS_DTLS0_9, GNUTLS_CIPHER_3DES_CBC, GNUTLS_MAC_SHA1,
-	  "NONE:+VERS-DTLS0.9:+COMP-NULL:+3DES-CBC:+SHA1:+RSA:%COMPAT" },
-#if GNUTLS_VERSION_NUMBER >= 0x030207 /* if DTLS 1.2 is supported (and a bug in gnutls is solved) */
-	{ "OC-DTLS1_2-AES128-GCM", GNUTLS_DTLS1_2, GNUTLS_CIPHER_AES_128_GCM, GNUTLS_MAC_AEAD,
-	  "NONE:+VERS-DTLS1.2:+COMP-NULL:+AES-128-GCM:+AEAD:+RSA:%COMPAT:+SIGN-ALL" },
-	{ "OC-DTLS1_2-AES256-GCM", GNUTLS_DTLS1_2, GNUTLS_CIPHER_AES_256_GCM, GNUTLS_MAC_AEAD,
-	  "NONE:+VERS-DTLS1.2:+COMP-NULL:+AES-256-GCM:+AEAD:+RSA:%COMPAT:+SIGN-ALL" },
-#endif
+	{ "AES128-SHA", GNUTLS_DTLS0_9, GNUTLS_CIPHER_AES_128_CBC, GNUTLS_KX_RSA, GNUTLS_MAC_SHA1,
+	  "NONE:+VERS-DTLS0.9:+COMP-NULL:+AES-128-CBC:+SHA1:+RSA:%COMPAT", "3.0.0" },
+	{ "AES256-SHA", GNUTLS_DTLS0_9, GNUTLS_CIPHER_AES_256_CBC, GNUTLS_KX_RSA, GNUTLS_MAC_SHA1,
+	  "NONE:+VERS-DTLS0.9:+COMP-NULL:+AES-256-CBC:+SHA1:+RSA:%COMPAT", "3.0.0" },
+	{ "DES-CBC3-SHA", GNUTLS_DTLS0_9, GNUTLS_CIPHER_3DES_CBC, GNUTLS_KX_RSA, GNUTLS_MAC_SHA1,
+	  "NONE:+VERS-DTLS0.9:+COMP-NULL:+3DES-CBC:+SHA1:+RSA:%COMPAT", "3.0.0" },
+	{ "OC-DTLS1_2-AES128-GCM", GNUTLS_DTLS1_2, GNUTLS_CIPHER_AES_128_GCM, GNUTLS_KX_RSA, GNUTLS_MAC_AEAD,
+	  "NONE:+VERS-DTLS1.2:+COMP-NULL:+AES-128-GCM:+AEAD:+RSA:%COMPAT:+SIGN-ALL", "3.2.7" },
+	{ "OC-DTLS1_2-AES256-GCM", GNUTLS_DTLS1_2, GNUTLS_CIPHER_AES_256_GCM, GNUTLS_KX_RSA, GNUTLS_MAC_AEAD,
+	  "NONE:+VERS-DTLS1.2:+COMP-NULL:+AES-256-GCM:+AEAD:+RSA:%COMPAT:+SIGN-ALL", "3.2.7" },
+	{ "OC2-DTLS1_2-CHACHA20-POLY1305", GNUTLS_DTLS1_2, GNUTLS_CIPHER_CHACHA20_POLY1305, GNUTLS_KX_PSK, GNUTLS_MAC_AEAD,
+	  "NONE:+VERS-DTLS1.2:+COMP-NULL:+CHACHA20-POLY1305:+AEAD:+PSK:%COMPAT:+SIGN-ALL", "3.4.8" },
+	/* NB. We agreed that any new cipher suites probably shouldn't use
+	 * Cisco's session resume hack (which ties us to a specific version
+	 * of DTLS). Instead, we'll use GNUTLS_KX_PSK and let it negotiate
+	 * the session properly. We might want to wait for
+	 * draft-jay-tls-psk-identity-extension before we do that. */
 };
 
+#if GNUTLS_VERSION_NUMBER < 0x030009
+void append_dtls_ciphers(struct openconnect_info *vpninfo, struct oc_text_buf *buf)
+{
+	int i, first = 1;
+
+	for (i = 0; i < sizeof(gnutls_dtls_ciphers) / sizeof(gnutls_dtls_ciphers[0]); i++) {
+		if (gnutls_check_version(gnutls_dtls_ciphers[i].min_gnutls_version)) {
+			buf_append(buf, "%s%s", first ? "" : ":",
+				   gnutls_dtls_ciphers[i].name);
+			first = 0;
+		}
+	}
+#else
+void append_dtls_ciphers(struct openconnect_info *vpninfo, struct oc_text_buf *buf)
+{
+	/* only enable the ciphers that would have been negotiated in the TLS channel */
+	unsigned i, j, first = 1;
+	int ret;
+	unsigned idx;
+	gnutls_cipher_algorithm_t cipher;
+	gnutls_mac_algorithm_t mac;
+	gnutls_priority_t cache;
+	uint32_t used = 0;
+
+	ret = gnutls_priority_init(&cache, vpninfo->gnutls_prio, NULL);
+	if (ret < 0) {
+		buf->error = -EIO;
+		return;
+	}
+
+	for (j=0; ; j++) {
+		ret = gnutls_priority_get_cipher_suite_index(cache, j, &idx);
+		if (ret == GNUTLS_E_UNKNOWN_CIPHER_SUITE)
+			continue;
+		else if (ret < 0)
+			break;
+
+		if (gnutls_cipher_suite_info(idx, NULL, NULL, &cipher, &mac, NULL) != NULL) {
+			for (i = 0; i < sizeof(gnutls_dtls_ciphers)/sizeof(gnutls_dtls_ciphers[0]); i++) {
+				if (used & (1 << i))
+					continue;
+				if (gnutls_dtls_ciphers[i].mac == mac && gnutls_dtls_ciphers[i].cipher == cipher) {
+					buf_append(buf, "%s%s", first ? "" : ":",
+						   gnutls_dtls_ciphers[i].name);
+					first = 0;
+					used |= (1 << i);
+					break;
+				}
+			}
+		}
+	}
+
+	gnutls_priority_deinit(cache);
+}
+#endif
+
 #define DTLS_SEND gnutls_record_send
 #define DTLS_RECV gnutls_record_recv
 #define DTLS_FREE gnutls_deinit
@@ -470,6 +567,8 @@ static int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd)
 	int cipher;
 
 	for (cipher = 0; cipher < sizeof(gnutls_dtls_ciphers)/sizeof(gnutls_dtls_ciphers[0]); cipher++) {
+		if (gnutls_check_version(gnutls_dtls_ciphers[cipher].min_gnutls_version) == NULL)
+			continue;
 		if (!strcmp(vpninfo->dtls_cipher, gnutls_dtls_ciphers[cipher].name))
 			goto found_cipher;
 	}
@@ -502,7 +601,7 @@ static int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd)
 	session_id.data = vpninfo->dtls_session_id;
 	session_id.size = sizeof(vpninfo->dtls_session_id);
 	err = gnutls_session_set_premaster(dtls_ssl, GNUTLS_CLIENT, gnutls_dtls_ciphers[cipher].version,
-					   GNUTLS_KX_RSA, gnutls_dtls_ciphers[cipher].cipher,
+					   gnutls_dtls_ciphers[cipher].kx, gnutls_dtls_ciphers[cipher].cipher,
 					   gnutls_dtls_ciphers[cipher].mac, GNUTLS_COMP_NULL,
 					   &master_secret, &session_id);
 	if (err) {
@@ -550,15 +649,22 @@ static int dtls_try_handshake(struct openconnect_info *vpninfo)
 		vpninfo->dtls_state = DTLS_CONNECTED;
 		str = get_gnutls_cipher(vpninfo->dtls_ssl);
 		if (str) {
+			const char *c;
 			vpn_progress(vpninfo, PRG_INFO,
 				     _("Established DTLS connection (using GnuTLS). Ciphersuite %s.\n"),
 				     str);
 			gnutls_free(str);
+			c = openconnect_get_dtls_compression(vpninfo);
+			if (c) {
+				vpn_progress(vpninfo, PRG_INFO,
+					     _("DTLS connection compression using %s.\n"), c);
+			}
 		}
 
 		vpninfo->dtls_times.last_rekey = vpninfo->dtls_times.last_rx = 
 			vpninfo->dtls_times.last_tx = time(NULL);
 
+		detect_mtu(vpninfo);
 		/* XXX: For OpenSSL we explicitly prevent retransmits here. */
 		return 0;
 	}
@@ -964,6 +1070,281 @@ int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout)
 
 	return work_done;
 }
+
+#if defined(DTLS_GNUTLS)
+
+static int is_cancelled(struct openconnect_info *vpninfo)
+{
+	fd_set rd_set;
+	int maxfd = 0;
+	FD_ZERO(&rd_set);
+	cmd_fd_set(vpninfo, &rd_set, &maxfd);
+
+	if (is_cancel_pending(vpninfo, &rd_set)) {
+		vpn_progress(vpninfo, PRG_ERR, _("SSL operation cancelled\n"));
+		return -EINTR;
+	}
+
+	return 0;
+}
+
+static
+void ms_sleep(unsigned ms)
+{
+	struct timespec tv;
+
+	tv.tv_sec = ms/1000;
+	tv.tv_nsec = (ms%1000) * 1000 * 1000;
+
+	nanosleep(&tv, NULL);
+}
+
+#define MTU_ID_SIZE 4
+#define MTU_FAIL_CONT { \
+		cur--; \
+		max = cur; \
+		continue; \
+	}
+
+/* Performs a binary search to detect MTU.
+ * @buf: is preallocated with MTU size
+ * @id: a unique ID for our DPD exchange
+ *
+ * Returns: new MTU or 0
+ */
+static int detect_mtu_ipv4(struct openconnect_info *vpninfo, unsigned char *buf)
+{
+	int max, min, cur, ret;
+	int max_tries = 100; /* maximum number of loops in bin search */
+	int max_recv_loops = 40; /* 4 secs max */
+	char id[MTU_ID_SIZE];
+
+	cur = max = vpninfo->ip_info.mtu;
+	min = vpninfo->ip_info.mtu/2;
+
+	vpn_progress(vpninfo, PRG_DEBUG,
+	     _("Initiating IPv4 MTU detection (min=%d, max=%d)\n"), min, max);
+
+	while (max > min) {
+		if (max_tries-- <= 0) {
+			vpn_progress(vpninfo, PRG_ERR,
+			     _("Too long time in MTU detect loop; bailing out.\n"));
+			goto fail;
+		}
+
+		/* generate unique ID */
+		if (gnutls_rnd(GNUTLS_RND_NONCE, id, sizeof(id)) < 0)
+			goto fail;
+
+		cur = (min + max + 1) / 2;
+		buf[0] = AC_PKT_DPD_OUT;
+		memcpy(&buf[1], id, sizeof(id));
+
+		do {
+			if (is_cancelled(vpninfo))
+				goto fail;
+
+			ret = DTLS_SEND(vpninfo->dtls_ssl, buf, cur+1);
+			if (ret < 0)
+				ms_sleep(100);
+		} while(ret == GNUTLS_E_AGAIN || ret == GNUTLS_E_INTERRUPTED);
+
+		if (ret != cur+1) {
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Failed to send DPD request (%d): %s\n"), cur, gnutls_strerror(ret));
+			MTU_FAIL_CONT;
+		}
+
+ reread:
+		memset(buf, 0, sizeof(id)+1);
+		do {
+			if (is_cancelled(vpninfo))
+				goto fail;
+
+			ret = DTLS_RECV(vpninfo->dtls_ssl, buf, cur + 1);
+			if (ret < 0)
+				ms_sleep(100);
+
+			if (max_recv_loops-- <= 0)
+				goto fail;
+		} while(ret == GNUTLS_E_AGAIN || ret == GNUTLS_E_INTERRUPTED);
+
+		if (ret > 0 && (buf[0] != AC_PKT_DPD_RESP || memcmp(&buf[1], id, sizeof(id)) != 0)) {
+			vpn_progress(vpninfo, PRG_DEBUG,
+			     _("Received unexpected packet (%.2x) in MTU detection; skipping.\n"), (unsigned)buf[0]);
+			goto reread; /* resend */
+		}
+
+		if (ret <= 0) {
+			vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to recv DPD request (%d): %s\n"), cur, gnutls_strerror(ret));
+			MTU_FAIL_CONT;
+		}
+
+		/* If we reached the max, success */
+		if (cur == max) {
+			break;
+		}
+		min = cur;
+	}
+
+	return cur;
+ fail:
+ 	return 0;
+}
+
+#if defined(IPPROTO_IPV6)
+
+/* This symbol is missing in glibc < 2.22 (bug 18643). */
+#if defined(__linux__) && !defined(HAVE_IPV6_PATHMTU)
+# define HAVE_IPV6_PATHMTU 1
+# define IPV6_PATHMTU 61
+#endif
+
+/* Verifies whether current MTU is ok, or detects new MTU using IPv6's ICMP6 messages
+ * @buf: is preallocated with MTU size
+ * @id: a unique ID for our DPD exchange
+ *
+ * Returns: new MTU or 0
+ */
+static int detect_mtu_ipv6(struct openconnect_info *vpninfo, unsigned char *buf)
+{
+	int max, cur, ret;
+	int max_resends = 5; /* maximum number of resends */
+	int max_recv_loops = 40; /* 4 secs max */
+	char id[MTU_ID_SIZE];
+
+	cur = max = vpninfo->ip_info.mtu;
+
+	vpn_progress(vpninfo, PRG_DEBUG,
+	     _("Initiating IPv6 MTU detection\n"));
+
+	do {
+		/* generate unique ID */
+		if (gnutls_rnd(GNUTLS_RND_NONCE, id, sizeof(id)) < 0)
+			goto fail;
+
+		buf[0] = AC_PKT_DPD_OUT;
+		memcpy(&buf[1], id, sizeof(id));
+		do {
+			if (is_cancelled(vpninfo))
+				goto fail;
+
+			ret = DTLS_SEND(vpninfo->dtls_ssl, buf, cur+1);
+			if (ret < 0)
+				ms_sleep(100);
+		} while(ret == GNUTLS_E_AGAIN || ret == GNUTLS_E_INTERRUPTED);
+
+		if (ret != cur+1) {
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Failed to send DPD request (%d): %s\n"), cur, gnutls_strerror(ret));
+			goto mtu6_fail;
+		}
+
+ reread:
+		memset(buf, 0, sizeof(id)+1);
+		do {
+			if (is_cancelled(vpninfo))
+				goto fail;
+
+			ret = DTLS_RECV(vpninfo->dtls_ssl, buf, cur + 1);
+			if (ret < 0)
+				ms_sleep(100);
+
+			if (max_recv_loops-- <= 0)
+				goto fail;
+		} while(ret == GNUTLS_E_AGAIN || ret == GNUTLS_E_INTERRUPTED);
+
+		/* timeout, probably our original request was lost,
+		 * let's resend the DPD */
+		if (ret == GNUTLS_E_TIMEDOUT)
+			continue;
+
+		/* something unexpected was received, let's ignore it */
+		if (ret > 0 && (buf[0] != AC_PKT_DPD_RESP || memcmp(&buf[1], id, sizeof(id)) != 0)) {
+			vpn_progress(vpninfo, PRG_DEBUG,
+			     _("Received unexpected packet (%.2x) in MTU detection; skipping.\n"), (unsigned)buf[0]);
+			goto reread;
+		}
+
+		/* we received what we expected, move on */
+		break;
+	} while(max_resends-- > 0);
+
+#ifdef HAVE_IPV6_PATHMTU
+	/* If we received back our DPD packet, do nothing; otherwise,
+	 * attempt to get MTU from the ICMP6 packet we received */
+	if (ret <= 0) {
+		struct ip6_mtuinfo mtuinfo;
+		socklen_t len = sizeof(mtuinfo);
+		max = 0;
+		vpn_progress(vpninfo, PRG_ERR,
+		     _("Failed to recv DPD request (%d): %s\n"), cur, gnutls_strerror(ret));
+ mtu6_fail:
+		if (getsockopt(vpninfo->dtls_fd, IPPROTO_IPV6, IPV6_PATHMTU, &mtuinfo, &len) >= 0) {
+			max = mtuinfo.ip6m_mtu;
+			if (max >= 0 && max < cur) {
+				gnutls_dtls_set_mtu(vpninfo->dtls_ssl, max);
+				cur = gnutls_dtls_get_data_mtu(vpninfo->dtls_ssl) - /*ipv6*/40 - /*udp*/20 - /*oc dtls*/1;
+			}
+		}
+	}
+#else
+ mtu6_fail:
+#endif
+
+	return cur;
+ fail:
+	return 0;
+}
+#endif
+
+static void detect_mtu(struct openconnect_info *vpninfo)
+{
+	int mtu = vpninfo->ip_info.mtu;
+	int prev_mtu = vpninfo->ip_info.mtu;
+	unsigned char *buf;
+
+	if (vpninfo->ip_info.mtu < 1+MTU_ID_SIZE)
+		return;
+
+	/* detect MTU */
+	buf = calloc(1, 1 + vpninfo->ip_info.mtu);
+	if (!buf) {
+		vpn_progress(vpninfo, PRG_ERR, _("Allocation failed\n"));
+		return;
+	}
+
+	/* enforce a timeout in receiving */
+	gnutls_record_set_timeout(vpninfo->dtls_ssl, 2400);
+	if (vpninfo->peer_addr->sa_family == AF_INET) { /* IPv4 */
+		mtu = detect_mtu_ipv4(vpninfo, buf);
+		if (mtu == 0)
+			goto skip_mtu;
+#if defined(IPPROTO_IPV6)
+	} else if (vpninfo->peer_addr->sa_family == AF_INET6) { /* IPv6 */
+		mtu = detect_mtu_ipv6(vpninfo, buf);
+		if (mtu == 0)
+			goto skip_mtu;
+#endif
+	}
+
+	vpninfo->ip_info.mtu = mtu;
+	if (prev_mtu != vpninfo->ip_info.mtu) {
+		vpn_progress(vpninfo, PRG_INFO,
+		     _("Detected MTU of %d bytes (was %d)\n"), vpninfo->ip_info.mtu, prev_mtu);
+	} else {
+		vpn_progress(vpninfo, PRG_DEBUG,
+		     _("No change in MTU after detection (was %d)\n"), prev_mtu);
+	}
+
+ skip_mtu:
+	gnutls_record_set_timeout(vpninfo->dtls_ssl, 0);
+	free(buf);
+}
+
+#endif
+
 #else /* !HAVE_DTLS */
 #warning Your SSL library does not seem to support Cisco DTLS compatibility
 #endif
