@@ -79,6 +79,25 @@ static P11KitPin *p11kit_pin_callback(const char *pin_source, P11KitUri *pin_uri
 #include "gnutls.h"
 #include "openconnect-internal.h"
 
+/* GnuTLS 2.x lacked this. But GNUTLS_E_UNEXPECTED_PACKET_LENGTH basically
+ * does the same thing.
+ * http://lists.infradead.org/pipermail/openconnect-devel/2014-March/001726.html
+ */
+#ifndef GNUTLS_E_PREMATURE_TERMINATION
+#define GNUTLS_E_PREMATURE_TERMINATION GNUTLS_E_UNEXPECTED_PACKET_LENGTH
+#endif
+
+
+/* Compile-time optimisable GnuTLS version check. We should never be
+ * run against a version of GnuTLS which is *older* than the one we
+ * were built again, but we might be run against a version which is
+ * newer. So some ancient compatibility code *can* be dropped at
+ * compile time. Likewise, if building against GnuTLS 2.x then we
+ * can never be running agsinst a 3.x library — the soname changed. */
+#define gtls_ver(a,b,c) ( GNUTLS_VERSION_MAJOR >= (a) &&		\
+	(GNUTLS_VERSION_NUMBER >= ( ((a) << 16) + ((b) << 8) + (c) ) || \
+	 gnutls_check_version(#a "." #b "." #c)))
+
 /* Helper functions for reading/writing lines over SSL. */
 static int openconnect_gnutls_write(struct openconnect_info *vpninfo, char *buf, size_t len)
 {
@@ -140,14 +159,16 @@ static int openconnect_gnutls_read(struct openconnect_info *vpninfo, char *buf,
 				vpn_progress(vpninfo, PRG_ERR, _("SSL read cancelled\n"));
 				return -EINTR;
 			}
-#ifdef GNUTLS_E_PREMATURE_TERMINATION
 		} else if (done == GNUTLS_E_PREMATURE_TERMINATION) {
 			/* We've seen this with HTTP 1.0 responses followed by abrupt
 			   socket closure and no clean SSL shutdown.
 			   https://bugs.launchpad.net/bugs/1225276 */
 			vpn_progress(vpninfo, PRG_DEBUG, _("SSL socket closed uncleanly\n"));
 			return 0;
-#endif
+		} else if (done == GNUTLS_E_REHANDSHAKE) {
+			int ret = cstp_handshake(vpninfo, 0);
+			if (ret)
+				return ret;
 		} else {
 			vpn_progress(vpninfo, PRG_ERR, _("Failed to read from SSL socket: %s\n"),
 				     gnutls_strerror(done));
@@ -203,6 +224,10 @@ static int openconnect_gnutls_gets(struct openconnect_info *vpninfo, char *buf,
 				ret = -EINTR;
 				break;
 			}
+		} else if (ret == GNUTLS_E_REHANDSHAKE) {
+			ret = cstp_handshake(vpninfo, 0);
+			if (ret)
+				return ret;
 		} else {
 			vpn_progress(vpninfo, PRG_ERR, _("Failed to read from SSL socket: %s\n"),
 				     gnutls_strerror(ret));
@@ -976,8 +1001,16 @@ static int load_certificate(struct openconnect_info *vpninfo)
 
 	key_is_p11 = !strncmp(vpninfo->sslkey, "pkcs11:", 7);
 	cert_is_p11 = !strncmp(vpninfo->cert, "pkcs11:", 7);
+
+#ifdef HAVE_GNUTLS_URL_IS_SUPPORTED
+	/* GnuTLS returns true for pkcs11:, tpmkey:, system:, and custom URLs. */
+	key_is_sys = !key_is_p11 && gnutls_url_is_supported(vpninfo->sslkey);
+	cert_is_sys = !cert_is_p11 && gnutls_url_is_supported(vpninfo->cert);
+#else
+	/* Fallback for GnuTLS < 3.1.0. */
 	key_is_sys = !strncmp(vpninfo->sslkey, "system:", 7);
 	cert_is_sys = !strncmp(vpninfo->cert, "system:", 7);
+#endif
 
 #ifndef HAVE_GNUTLS_SYSTEM_KEYS
 	if (key_is_sys || cert_is_sys) {
@@ -1574,7 +1607,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 		err = gnutls_certificate_set_x509_crl(vpninfo->https_cred, &crl, 1);
 		if (err) {
 			vpn_progress(vpninfo, PRG_ERR,
-				     _("Setting certificate recovation list failed: %s\n"),
+				     _("Setting certificate revocation list failed: %s\n"),
 				     gnutls_strerror(err));
 			ret = -EINVAL;
 			goto out;
@@ -1923,6 +1956,38 @@ void openconnect_free_cert_info(struct openconnect_info *vpninfo,
 	gnutls_free(buf);
 }
 
+int openconnect_get_peer_cert_chain(struct openconnect_info *vpninfo,
+				    struct oc_cert **chainp)
+{
+	struct oc_cert *chain, *p;
+	const gnutls_datum_t *cert_list = vpninfo->cert_list_handle;
+	int i, cert_list_size = vpninfo->cert_list_size;
+
+	if (!cert_list)
+		return -EINVAL;
+
+	if (cert_list_size <= 0)
+		return -EIO;
+
+	p = chain = calloc(cert_list_size, sizeof(struct oc_cert));
+	if (!chain)
+		return -ENOMEM;
+
+	for (i = 0; i < cert_list_size; i++, p++) {
+		p->der_data = (unsigned char *)cert_list[i].data;
+		p->der_len = cert_list[i].size;
+	}
+
+	*chainp = chain;
+	return cert_list_size;
+}
+
+void openconnect_free_peer_cert_chain(struct openconnect_info *vpninfo,
+				      struct oc_cert *chain)
+{
+	free(chain);
+}
+
 static int verify_peer(gnutls_session_t session)
 {
 	struct openconnect_info *vpninfo = gnutls_session_get_ptr(session);
@@ -1938,6 +2003,24 @@ static int verify_peer(gnutls_session_t session)
 		return GNUTLS_E_CERTIFICATE_ERROR;
 	}
 
+	if (vpninfo->peer_cert) {
+		unsigned char *prev_der = NULL;
+		int der_len = openconnect_get_peer_cert_DER(vpninfo, &prev_der);
+		if (der_len < 0) {
+			vpn_progress(vpninfo, PRG_ERR, _("Error comparing server's cert on rehandshake: %s\n"),
+				     strerror(-der_len));
+			return GNUTLS_E_CERTIFICATE_ERROR;
+		}
+		if (cert_list[0].size != der_len || memcmp(cert_list[0].data, prev_der, der_len)) {
+			vpn_progress(vpninfo, PRG_ERR, _("Server presented different cert on rehandshake\n"));
+			gnutls_free(prev_der);
+			return GNUTLS_E_CERTIFICATE_ERROR;
+		}
+		gnutls_free(prev_der);
+		vpn_progress(vpninfo, PRG_TRACE, _("Server presented identical cert on rehandshake\n"));
+		return 0;
+	}
+
 	err = gnutls_x509_crt_init(&cert);
 	if (err) {
 		vpn_progress(vpninfo, PRG_ERR, _("Error initialising X509 cert structure\n"));
@@ -2008,6 +2091,7 @@ static int verify_peer(gnutls_session_t session)
 		else if (inet_pton(AF_INET6, vpninfo->hostname, addrbuf) > 0)
 			addrlen = 16;
 #endif
+
 		if (!addrlen) {
 			/* vpninfo->hostname was not a bare IP address. Nothing to do */
 			goto badhost;
@@ -2035,40 +2119,24 @@ static int verify_peer(gnutls_session_t session)
 		vpn_progress(vpninfo, PRG_INFO,
 			     _("Server certificate verify failed: %s\n"),
 			     reason);
-		if (vpninfo->validate_peer_cert)
+		if (vpninfo->validate_peer_cert) {
+			vpninfo->cert_list_handle = (void *)cert_list;
+			vpninfo->cert_list_size = cert_list_size;
 			err = vpninfo->validate_peer_cert(vpninfo->cbdata,
 							  reason) ? GNUTLS_E_CERTIFICATE_ERROR : 0;
-		else
+			vpninfo->cert_list_handle = NULL;
+		} else
 			err = GNUTLS_E_CERTIFICATE_ERROR;
 	}
 
 	return err;
 }
 
-
-/* The F5 firewall is confused when the TLS client hello is between
- * 256 and 512 bytes. By disabling several TLS options we force the
- * client hello to be < 256 bytes. We don't do that in gnutls versions
- * >= 3.2.9 as there the %COMPAT keyword ensures that the client hello
- * will be outside that range.
- */
-#if GNUTLS_VERSION_NUMBER >= 0x030209
-# define DEFAULT_PRIO "NORMAL:-VERS-SSL3.0:%COMPAT"
-#else
-# define _DEFAULT_PRIO "NORMAL:-VERS-TLS-ALL:+VERS-TLS1.0:" \
-	"%COMPAT:%DISABLE_SAFE_RENEGOTIATION:%LATEST_RECORD_VERSION"
-# if GNUTLS_VERSION_MAJOR >= 3
-#  define DEFAULT_PRIO _DEFAULT_PRIO":-CURVE-ALL:-ECDHE-RSA:-ECDHE-ECDSA"
-#else
-#  define DEFAULT_PRIO _DEFAULT_PRIO
-# endif
-#endif
-
 int openconnect_open_https(struct openconnect_info *vpninfo)
 {
+	const char *default_prio;
 	int ssl_sock = -1;
 	int err;
-	const char * prio;
 
 	if (vpninfo->https_sess)
 		return 0;
@@ -2078,7 +2146,7 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 		vpninfo->peer_cert = NULL;
 	}
 	free(vpninfo->peer_cert_hash);
-	vpninfo->peer_cert_hash = 0;
+	vpninfo->peer_cert_hash = NULL;
 	gnutls_free(vpninfo->cstp_cipher);
 	vpninfo->cstp_cipher = NULL;
 
@@ -2185,26 +2253,65 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 	if (vpninfo->my_pkey == OPENCONNECT_TPM_PKEY)
 		gnutls_sign_callback_set(vpninfo->https_sess, gtls2_tpm_sign_cb, vpninfo);
 #endif
-	/* We depend on 3.2.9 because that has the workaround for the
-	   obnoxious F5 firewall that drops packets of certain sizes */
-	if (gnutls_check_version("3.2.9") &&
-	    string_is_hostname(vpninfo->hostname))
+	/*
+	 * For versions of GnuTLS older than 3.2.9, we try to avoid long
+	 * packets by silently disabling extensions such as SNI.
+	 *
+	 * See comments above regarding COMPAT and DUMBFW.
+	 */
+	if (gtls_ver(3,2,9) && string_is_hostname(vpninfo->hostname))
 		gnutls_server_name_set(vpninfo->https_sess, GNUTLS_NAME_DNS,
 				       vpninfo->hostname,
 				       strlen(vpninfo->hostname));
 
-	if (vpninfo->pfs) {
-		prio = DEFAULT_PRIO":-RSA";
+       /*
+	* If a ClientHello is between 256 and 511 bytes, the
+	* server cannot distinguish between a SSLv2 formatted
+	* packet and a SSLv3 formatted packet.
+	*
+	* F5 BIG-IP reverse proxies in particular will
+	* silently drop an ambiguous ClientHello.
+	*
+	* GnuTLS fixes this in v3.2.9+ by padding ClientHello
+	* packets to at least 512 bytes if %COMPAT or %DUMBFW
+	* is specified.
+	*
+	* Discussion:
+	* http://www.ietf.org/mail-archive/web/tls/current/msg10423.html
+	*
+	* GnuTLS commits:
+	* b6d29bb1737f96ac44a8ef9cc9fe7f9837e20465
+	* a9bd8c4d3a639c40adb964349297f891f583a21b
+	* 531bec47037e882af32963f8461988f8c724919e
+	* 7c45ebbdd877cd994b6b938bd6faef19558a01e1
+	* 8d28901a3ebd2589d0fc9941475d50f04047f6fe
+	* 28065ce3896b1b0f87972d0bce9b17641ebb69b9
+	*/
+
+#ifdef DEFAULT_PRIO
+	default_prio = DEFAULT_PRIO ":%COMPAT";
+#else
+	if (gtls_ver(3,2,9)) {
+		default_prio = "NORMAL:-VERS-SSL3.0:%COMPAT";
+	} else if (gtls_ver(3,0,0)) {
+		default_prio = "NORMAL:-VERS-TLS-ALL:+VERS-TLS1.0:" \
+			"%COMPAT:%DISABLE_SAFE_RENEGOTIATION:%LATEST_RECORD_VERSION" \
+			":-CURVE-ALL:-ECDHE-RSA:-ECDHE-ECDSA";
 	} else {
-		prio = DEFAULT_PRIO;
+		default_prio = "NORMAL:-VERS-TLS-ALL:+VERS-TLS1.0:"			\
+			"%COMPAT:%DISABLE_SAFE_RENEGOTIATION:%LATEST_RECORD_VERSION";
 	}
+#endif
+
+	snprintf(vpninfo->gnutls_prio, sizeof(vpninfo->gnutls_prio), "%s%s",
+		 default_prio, vpninfo->pfs?":-RSA":"");
 
 	err = gnutls_priority_set_direct(vpninfo->https_sess,
-					prio, NULL);
+					 vpninfo->gnutls_prio, NULL);
 	if (err) {
 		vpn_progress(vpninfo, PRG_ERR,
-			     _("Failed to set TLS priority string: %s\n"),
-			     gnutls_strerror(err));
+			     _("Failed to set TLS priority string (\"%s\"): %s\n"),
+			     vpninfo->gnutls_prio, gnutls_strerror(err));
 		gnutls_deinit(vpninfo->https_sess);
 		vpninfo->https_sess = NULL;
 		closesocket(ssl_sock);
