@@ -104,43 +104,53 @@ err:
 	return NULL;
 }
 
+const struct vpn_proto openconnect_protos[] = {
+	{
+		.name = "anyconnect",
+		.vpn_close_session = cstp_bye,
+		.tcp_connect = cstp_connect,
+		.tcp_mainloop = cstp_mainloop,
+		.add_http_headers = cstp_common_headers,
+		.obtain_cookie = cstp_obtain_cookie,
+#ifdef HAVE_DTLS
+		.udp_setup = dtls_setup,
+		.udp_mainloop = dtls_mainloop,
+		.udp_close = dtls_close,
+		.udp_shutdown = dtls_shutdown,
+#endif
+	}, {
+		.name = "nc",
+		.vpn_close_session = NULL,
+		.tcp_connect = oncp_connect,
+		.tcp_mainloop = oncp_mainloop,
+		.add_http_headers = oncp_common_headers,
+		.obtain_cookie = oncp_obtain_cookie,
+#if defined(ESP_GNUTLS) || defined(ESP_OPENSSL)
+		.udp_setup = esp_setup,
+		.udp_mainloop = esp_mainloop,
+		.udp_close = esp_close,
+		.udp_shutdown = esp_shutdown,
+#endif
+	},
+	{ /* NULL */ }
+};
+
 int openconnect_set_protocol(struct openconnect_info *vpninfo, const char *protocol)
 {
-	if (!strcmp(protocol, "anyconnect")) {
-		vpninfo->proto.vpn_close_session = cstp_bye;
-		vpninfo->proto.tcp_connect = cstp_connect;
-		vpninfo->proto.tcp_mainloop = cstp_mainloop;
-		vpninfo->proto.add_http_headers = cstp_common_headers;
-		vpninfo->proto.obtain_cookie = cstp_obtain_cookie;
-#ifdef HAVE_DTLS
-		vpninfo->proto.udp_setup = dtls_setup;
-		vpninfo->proto.udp_mainloop = dtls_mainloop;
-		vpninfo->proto.udp_close = dtls_close;
-		vpninfo->proto.udp_shutdown = dtls_shutdown;
-#else
-		vpninfo->dtls_state = DTLS_DISABLED;
-#endif
-	} else if (!strcmp(protocol, "nc")) {
-		vpninfo->proto.vpn_close_session = NULL;
-		vpninfo->proto.tcp_connect = oncp_connect;
-		vpninfo->proto.tcp_mainloop = oncp_mainloop;
-		vpninfo->proto.add_http_headers = oncp_common_headers;
-		vpninfo->proto.obtain_cookie = oncp_obtain_cookie;
-#if defined(ESP_GNUTLS) || defined(ESP_OPENSSL)
-		vpninfo->proto.udp_setup = esp_setup;
-		vpninfo->proto.udp_mainloop = esp_mainloop;
-		vpninfo->proto.udp_close = esp_close;
-		vpninfo->proto.udp_shutdown = esp_shutdown;
-#else
-		vpninfo->dtls_state = DTLS_DISABLED;
-#endif
-	} else {
-		vpn_progress(vpninfo, PRG_ERR,
-			     _("Unknown VPN protocol '%s'\n"), protocol);
-		return -EINVAL;
-	}
+	const struct vpn_proto *p;
 
-	return 0;
+	for (p = openconnect_protos; p->name; p++) {
+		if (strcasecmp(p->name, protocol))
+			continue;
+		vpninfo->proto = p;
+		if (!p->udp_setup)
+			vpninfo->dtls_state = DTLS_DISABLED;
+
+		return 0;
+	}
+	vpn_progress(vpninfo, PRG_ERR,
+		     _("Unknown VPN protocol '%s'\n"), protocol);
+	return -EINVAL;
 }
 
 void openconnect_set_loglevel(struct openconnect_info *vpninfo,
@@ -153,8 +163,8 @@ int openconnect_setup_dtls(struct openconnect_info *vpninfo,
 			   int attempt_period)
 
 {
-	if (vpninfo->proto.udp_setup)
-		return vpninfo->proto.udp_setup(vpninfo, attempt_period);
+	if (vpninfo->proto->udp_setup)
+		return vpninfo->proto->udp_setup(vpninfo, attempt_period);
 
 	vpn_progress(vpninfo, PRG_ERR,
 		     _("Built against SSL library with no Cisco DTLS support\n"));
@@ -163,12 +173,12 @@ int openconnect_setup_dtls(struct openconnect_info *vpninfo,
 
 int openconnect_obtain_cookie(struct openconnect_info *vpninfo)
 {
-	return vpninfo->proto.obtain_cookie(vpninfo);
+	return vpninfo->proto->obtain_cookie(vpninfo);
 }
 
 int openconnect_make_cstp_connection(struct openconnect_info *vpninfo)
 {
-	return vpninfo->proto.tcp_connect(vpninfo);
+	return vpninfo->proto->tcp_connect(vpninfo);
 }
 
 int openconnect_set_reported_os(struct openconnect_info *vpninfo,
@@ -227,8 +237,8 @@ static void free_optlist(struct oc_vpn_option *opt)
 void openconnect_vpninfo_free(struct openconnect_info *vpninfo)
 {
 	openconnect_close_https(vpninfo, 1);
-	if (vpninfo->proto.udp_shutdown)
-		vpninfo->proto.udp_shutdown(vpninfo);
+	if (vpninfo->proto->udp_shutdown)
+		vpninfo->proto->udp_shutdown(vpninfo);
 	if (vpninfo->tncc_fd != -1)
 		closesocket(vpninfo->tncc_fd);
 	if (vpninfo->cmd_fd_write != -1) {
@@ -252,6 +262,7 @@ void openconnect_vpninfo_free(struct openconnect_info *vpninfo)
 		CloseHandle(vpninfo->dtls_event);
 #endif
 	free(vpninfo->peer_addr);
+	free(vpninfo->ip_info.gateway_addr);
 	free_optlist(vpninfo->csd_env);
 	free_optlist(vpninfo->script_env);
 	free_optlist(vpninfo->cookies);
@@ -375,6 +386,11 @@ const char *openconnect_get_hostname(struct openconnect_info *vpninfo)
 	return vpninfo->unique_hostname?:vpninfo->hostname;
 }
 
+const char *openconnect_get_dnsname(struct openconnect_info *vpninfo)
+{
+	return vpninfo->hostname;
+}
+
 int openconnect_set_hostname(struct openconnect_info *vpninfo,
 			     const char *hostname)
 {
@@ -385,6 +401,8 @@ int openconnect_set_hostname(struct openconnect_info *vpninfo,
 	vpninfo->unique_hostname = NULL;
 	free(vpninfo->peer_addr);
 	vpninfo->peer_addr = NULL;
+	free(vpninfo->ip_info.gateway_addr);
+	vpninfo->ip_info.gateway_addr = NULL;
 
 	return 0;
 }
@@ -403,6 +421,15 @@ int openconnect_set_urlpath(struct openconnect_info *vpninfo,
 	return 0;
 }
 
+int openconnect_set_localname(struct openconnect_info *vpninfo,
+			      const char *localname)
+{
+	UTF8CHECK(localname);
+
+	STRDUP(vpninfo->localname, localname);
+	return 0;
+}
+
 void openconnect_set_xmlsha1(struct openconnect_info *vpninfo,
 			     const char *xmlsha1, int size)
 {
@@ -412,6 +439,11 @@ void openconnect_set_xmlsha1(struct openconnect_info *vpninfo,
 	memcpy(&vpninfo->xmlsha1, xmlsha1, size);
 }
 
+void openconnect_disable_ipv6(struct openconnect_info *vpninfo)
+{
+	vpninfo->disable_ipv6 = 1;
+}
+
 int openconnect_set_cafile(struct openconnect_info *vpninfo, const char *cafile)
 {
 	UTF8CHECK(cafile);
@@ -461,8 +493,10 @@ int openconnect_get_ip_info(struct openconnect_info *vpninfo,
 int openconnect_setup_csd(struct openconnect_info *vpninfo, uid_t uid,
 			  int silent, const char *wrapper)
 {
+#ifndef _WIN32
 	vpninfo->uid_csd = uid;
 	vpninfo->uid_csd_given = silent ? 2 : 1;
+#endif
 	STRDUP(vpninfo->csd_wrapper, wrapper);
 
 	return 0;
@@ -514,10 +548,12 @@ void openconnect_reset_ssl(struct openconnect_info *vpninfo)
 {
 	vpninfo->got_cancel_cmd = 0;
 	openconnect_close_https(vpninfo, 0);
-	if (vpninfo->peer_addr) {
-		free(vpninfo->peer_addr);
-		vpninfo->peer_addr = NULL;
-	}
+
+	free(vpninfo->peer_addr);
+	vpninfo->peer_addr = NULL;
+	free(vpninfo->ip_info.gateway_addr);
+	vpninfo->ip_info.gateway_addr = NULL;
+
 	openconnect_clear_cookies(vpninfo);
 }
 
@@ -754,6 +790,23 @@ void openconnect_set_protect_socket_handler(struct openconnect_info *vpninfo,
 	vpninfo->protect_socket = protect_socket;
 }
 
+void openconnect_override_getaddrinfo(struct openconnect_info *vpninfo, openconnect_getaddrinfo_vfn gai_fn)
+{
+	vpninfo->getaddrinfo_override = gai_fn;
+}
+
+void openconnect_set_setup_tun_handler(struct openconnect_info *vpninfo,
+				       openconnect_setup_tun_vfn setup_tun)
+{
+	vpninfo->setup_tun = setup_tun;
+}
+
+void openconnect_set_reconnected_handler(struct openconnect_info *vpninfo,
+				         openconnect_reconnected_vfn reconnected)
+{
+	vpninfo->reconnected = reconnected;
+}
+
 void openconnect_set_stats_handler(struct openconnect_info *vpninfo,
 				   openconnect_stats_vfn stats_handler)
 {
@@ -789,6 +842,28 @@ int openconnect_setup_tun_device(struct openconnect_info *vpninfo,
 	return openconnect_setup_tun_fd(vpninfo, tun_fd);
 }
 
+static const char *compr_name_map[] = {
+	[COMPR_DEFLATE] = "Deflate",
+	[COMPR_LZS] = "LZS",
+	[COMPR_LZ4] = "LZ4"
+};
+
+const char *openconnect_get_cstp_compression(struct openconnect_info * vpninfo)
+{
+	if (vpninfo->cstp_compr <= 0 || vpninfo->cstp_compr > COMPR_MAX)
+		return NULL;
+
+	return compr_name_map[vpninfo->cstp_compr];
+}
+
+const char *openconnect_get_dtls_compression(struct openconnect_info * vpninfo)
+{
+	if (vpninfo->dtls_compr <= 0 || vpninfo->dtls_compr > COMPR_MAX)
+		return NULL;
+
+	return compr_name_map[vpninfo->dtls_compr];
+}
+
 const char *openconnect_get_dtls_cipher(struct openconnect_info *vpninfo)
 {
 #if defined(DTLS_GNUTLS)
