@@ -85,6 +85,7 @@ static int do_passphrase_from_fsid;
 static int nocertcheck;
 static int non_inter;
 static int cookieonly;
+static int allow_stdin_read;
 
 static char *token_filename;
 static char *server_cert = NULL;
@@ -122,12 +123,12 @@ static void __attribute__ ((format(printf, 3, 4)))
 	}
 }
 #define openlog(...)  /* */
-#elif defined(_WIN32)
+#elif defined(_WIN32) || defined(__native_client__)
 /*
  * FIXME: Perhaps we could implement syslog_progress() using these APIs:
  * http://msdn.microsoft.com/en-us/library/windows/desktop/aa364148%28v=vs.85%29.aspx
  */
-#else /* !__ANDROID__ && !_WIN32 */
+#else /* !__ANDROID__ && !_WIN32 && !__native_client__ */
 #include <syslog.h>
 static void  __attribute__ ((format(printf, 3, 4)))
     syslog_progress(void *_vpninfo, int level, const char *fmt, ...)
@@ -174,6 +175,7 @@ enum {
 	OPT_PRINTCOOKIE,
 	OPT_RECONNECT_TIMEOUT,
 	OPT_SERVERCERT,
+	OPT_RESOLVE,
 	OPT_USERAGENT,
 	OPT_NON_INTER,
 	OPT_DTLS_LOCAL_PORT,
@@ -184,6 +186,8 @@ enum {
 	OPT_PFS,
 	OPT_PROXY_AUTH,
 	OPT_HTTP_AUTH,
+	OPT_LOCAL_HOSTNAME,
+	OPT_PROTOCOL,
 };
 
 #ifdef __sun__
@@ -245,8 +249,10 @@ static const struct option long_options[] = {
 	OPTION("dtls-ciphers", 1, OPT_DTLS_CIPHERS),
 	OPTION("authgroup", 1, OPT_AUTHGROUP),
 	OPTION("servercert", 1, OPT_SERVERCERT),
+	OPTION("resolve", 1, OPT_RESOLVE),
 	OPTION("key-password-from-fsid", 0, OPT_KEY_PASSWORD_FROM_FSID),
 	OPTION("useragent", 1, OPT_USERAGENT),
+	OPTION("local-hostname", 1, OPT_LOCAL_HOSTNAME),
 	OPTION("disable-ipv6", 0, OPT_DISABLE_IPV6),
 	OPTION("no-proxy", 0, OPT_NO_PROXY),
 	OPTION("libproxy", 0, OPT_LIBPROXY),
@@ -261,6 +267,7 @@ static const struct option long_options[] = {
 	OPTION("no-xmlpost", 0, OPT_NO_XMLPOST),
 	OPTION("dump-http-traffic", 0, OPT_DUMP_HTTP),
 	OPTION("no-system-trust", 0, OPT_NO_SYSTEM_TRUST),
+	OPTION("protocol", 1, OPT_PROTOCOL),
 #ifdef OPENCONNECT_GNUTLS
 	OPTION("gnutls-debug", 1, OPT_GNUTLS_DEBUG),
 #endif
@@ -356,7 +363,7 @@ static char *convert_arg_to_utf8(char **argv, char *arg)
 #define vfprintf vfprintf_utf8
 #define is_arg_utf8(str) (0)
 
-static void read_stdin(char **string, int hidden)
+static void read_stdin(char **string, int hidden, int allow_fail)
 {
 	CONSOLE_READCONSOLE_CONTROL rcc = { sizeof(rcc), 0, 13, 0 };
 	HANDLE stdinh = GetStdHandle(STD_INPUT_HANDLE);
@@ -373,6 +380,7 @@ static void read_stdin(char **string, int hidden)
 		char *errstr = openconnect__win32_strerror(GetLastError());
 		fprintf(stderr, _("ReadConsole() failed: %s\n"), errstr);
 		free(errstr);
+		*string = NULL;
 		goto out;
 	}
 
@@ -620,7 +628,7 @@ static void print_build_opts(void)
 
 #ifndef _WIN32
 static const char default_vpncscript[] = DEFAULT_VPNCSCRIPT;
-static void read_stdin(char **string, int hidden)
+static void read_stdin(char **string, int hidden, int allow_fail)
 {
 	char *c, *buf = malloc(1025);
 	int fd = fileno(stdin);
@@ -646,8 +654,14 @@ static void read_stdin(char **string, int hidden)
 	}
 
 	if (!buf) {
-		perror(_("fgets (stdin)"));
-		exit(1);
+		if (allow_fail) {
+			*string = NULL;
+			free(buf);
+			return;
+		} else {
+			perror(_("fgets (stdin)"));
+			exit(1);
+		}
 	}
 
 	c = strchr(buf, '\n');
@@ -700,6 +714,28 @@ static void set_default_vpncscript(void)
 }
 #endif
 
+static struct oc_vpn_option *gai_overrides;
+
+static int gai_override_cb(void *cbdata, const char *node,
+			    const char *service, const struct addrinfo *hints,
+			    struct addrinfo **res)
+{
+	struct openconnect_info *vpninfo = cbdata;
+	struct oc_vpn_option *p = gai_overrides;
+
+	while (p) {
+		if (!strcmp(node, p->option)) {
+			vpn_progress(vpninfo, PRG_TRACE, _("Override hostname '%s' to '%s'\n"),
+				     node, p->value);
+			node = p->value;
+			break;
+		}
+		p = p->next;
+	}
+
+	return getaddrinfo(node, service, hints, res);
+}
+
 static void usage(void)
 {
 	printf(_("Usage:  openconnect [options] <server>\n"));
@@ -780,6 +816,8 @@ static void usage(void)
 	printf("      --reconnect-timeout         %s\n", _("Connection retry timeout in seconds"));
 	printf("      --servercert=FINGERPRINT    %s\n", _("Server's certificate SHA1 fingerprint"));
 	printf("      --useragent=STRING          %s\n", _("HTTP header User-Agent: field"));
+	printf("      --local-hostname=STRING     %s\n", _("Local hostname to advertise to server"));
+	printf("      --resolve=HOST:IP           %s\n", _("Use IP when connecting to HOST"));
 	printf("      --os=STRING                 %s\n", _("OS type (linux,linux-64,win,...) to report"));
 	printf("      --dtls-local-port=PORT      %s\n", _("Set local port for DTLS datagrams"));
 	printf("\n");
@@ -930,13 +968,46 @@ static int next_option(int argc, char **argv, char **config_arg)
 
 }
 
+#ifndef _WIN32
+static void get_uids(const char *config_arg, uid_t *uid, gid_t *gid)
+{
+	char *strend;
+	struct passwd *pw;
+	int e;
+
+	*uid = strtol(config_arg, &strend, 0);
+	if (strend[0]) {
+		pw = getpwnam(config_arg);
+		if (!pw) {
+			e = errno;
+			fprintf(stderr, _("Invalid user \"%s\": %s\n"),
+				config_arg, strerror(e));
+			exit(1);
+		}
+		*uid = pw->pw_uid;
+		*gid = pw->pw_gid;
+	} else {
+		pw = getpwuid(*uid);
+		if (!pw) {
+			e = errno;
+			fprintf(stderr, _("Invalid user ID \"%d\": %s\n"),
+				(int)*uid, strerror(e));
+			exit(1);
+		}
+		*gid = pw->pw_gid;
+	}
+}
+#endif
+
 int main(int argc, char **argv)
 {
 	struct openconnect_info *vpninfo;
 	char *urlpath = NULL;
+	struct oc_vpn_option *gai;
+	char *ip;
 	const char *compr = "";
 	char *proxy = getenv("https_proxy");
-	char *vpnc_script = NULL, *ifname = NULL;
+	char *vpnc_script = NULL;
 	const struct oc_ip_info *ip_info;
 	int autoproxy = 0;
 	int opt;
@@ -954,9 +1025,7 @@ int main(int argc, char **argv)
 #ifndef _WIN32
 	struct sigaction sa;
 	struct utsname utsbuf;
-	uid_t uid = getuid();
 	int use_syslog = 0;
-	int script_tun = 0;
 #endif
 
 #ifdef ENABLE_NLS
@@ -998,9 +1067,12 @@ int main(int argc, char **argv)
 #ifdef _WIN32
 	set_default_vpncscript();
 #else
+	vpninfo->use_tun_script = 0;
+	vpninfo->uid = getuid();
+	vpninfo->gid = getgid();
+
 	if (!uname(&utsbuf)) {
-		free(vpninfo->localname);
-		vpninfo->localname = xstrdup(utsbuf.nodename);
+		openconnect_set_localname(vpninfo, utsbuf.nodename);
 	}
 #endif
 
@@ -1018,46 +1090,28 @@ int main(int argc, char **argv)
 			use_syslog = 1;
 			break;
 		case 'S':
-			script_tun = 1;
+			vpninfo->use_tun_script = 1;
 			break;
-		case 'U': {
-			char *strend;
-			uid = strtol(config_arg, &strend, 0);
-			if (strend[0]) {
-				struct passwd *pw = getpwnam(config_arg);
-				if (!pw) {
-					fprintf(stderr, _("Invalid user \"%s\"\n"),
-						config_arg);
-					exit(1);
-				}
-				uid = pw->pw_uid;
-			}
+		case 'U':
+			get_uids(config_arg, &vpninfo->uid, &vpninfo->gid);
+			break;
+		case OPT_CSD_USER:
+			get_uids(config_arg, &vpninfo->uid_csd, &vpninfo->gid_csd);
+			vpninfo->uid_csd_given = 1;
+			break;
+		case OPT_CSD_WRAPPER:
+			vpninfo->csd_wrapper = keep_config_arg();
+			break;
+#endif /* !_WIN32 */
+		case OPT_PROTOCOL:
+			if (openconnect_set_protocol(vpninfo, config_arg))
+				exit(1);
 			break;
-		}
 		case OPT_JUNIPER:
 			fprintf(stderr, "WARNING: Juniper Network Connect support is experimental.\n");
 			fprintf(stderr, "It will probably be superseded by Junos Pulse support.\n");
 			openconnect_set_protocol(vpninfo, "nc");
 			break;
-		case OPT_CSD_USER: {
-			char *strend;
-			vpninfo->uid_csd = strtol(config_arg, &strend, 0);
-			if (strend[0]) {
-				struct passwd *pw = getpwnam(config_arg);
-				if (!pw) {
-					fprintf(stderr, _("Invalid user \"%s\"\n"),
-						config_arg);
-					exit(1);
-				}
-				vpninfo->uid_csd = pw->pw_uid;
-			}
-			vpninfo->uid_csd_given = 1;
-			break;
-		}
-		case OPT_CSD_WRAPPER:
-			vpninfo->csd_wrapper = keep_config_arg();
-			break;
-#endif /* !_WIN32 */
 		case OPT_CONFIGFILE:
 			if (config_file) {
 				fprintf(stderr, _("Cannot use 'config' option inside config file\n"));
@@ -1102,6 +1156,24 @@ int main(int argc, char **argv)
 			server_cert = keep_config_arg();
 			openconnect_set_system_trust(vpninfo, 0);
 			break;
+		case OPT_RESOLVE:
+			ip = strchr(config_arg, ':');
+			if (!ip) {
+				fprintf(stderr, _("Missing colon in resolve option\n"));
+				exit(1);
+			}
+			gai = malloc(sizeof(*gai) + strlen(config_arg));
+			if (!gai) {
+				fprintf(stderr, _("Failed to allocate memory\n"));
+				exit(1);
+			}
+			gai->next = gai_overrides;
+			gai_overrides = gai;
+			gai->option = (void *)(gai + 1);
+			memcpy(gai->option, config_arg, strlen(config_arg));
+			gai->option[ip - config_arg] = 0;
+			gai->value = gai->option + (ip - config_arg) + 1;
+			break;
 		case OPT_NO_DTLS:
 			vpninfo->dtls_state = DTLS_DISABLED;
 			break;
@@ -1115,13 +1187,14 @@ int main(int argc, char **argv)
 			cookieonly = 3;
 			break;
 		case OPT_COOKIE_ON_STDIN:
-			read_stdin(&vpninfo->cookie, 0);
+			read_stdin(&vpninfo->cookie, 0, 0);
 			/* If the cookie is empty, ignore it */
 			if (!*vpninfo->cookie)
 				vpninfo->cookie = NULL;
 			break;
 		case OPT_PASSWORD_ON_STDIN:
-			read_stdin(&password, 0);
+			read_stdin(&password, 0, 0);
+			allow_stdin_read = 1;
 			break;
 		case OPT_NO_PASSWD:
 			vpninfo->nopasswd = 1;
@@ -1166,7 +1239,7 @@ int main(int argc, char **argv)
 		case 'h':
 			usage();
 		case 'i':
-			ifname = dup_config_arg();
+			vpninfo->ifname = dup_config_arg();
 			break;
 		case 'm': {
 			int mtu = atol(config_arg);
@@ -1258,6 +1331,9 @@ int main(int argc, char **argv)
 			free(vpninfo->useragent);
 			vpninfo->useragent = dup_config_arg();
 			break;
+		case OPT_LOCAL_HOSTNAME:
+			openconnect_set_localname(vpninfo, config_arg);
+			break;
 		case OPT_FORCE_DPD:
 			openconnect_set_dpd(vpninfo, atoi(config_arg));
 			break;
@@ -1310,6 +1386,9 @@ int main(int argc, char **argv)
 		}
 	}
 
+	if (gai_overrides)
+		openconnect_override_getaddrinfo(vpninfo, gai_override_cb);
+
 	if (optind < argc - 1) {
 		fprintf(stderr, _("Too many arguments on command line\n"));
 		usage();
@@ -1341,12 +1420,14 @@ int main(int argc, char **argv)
 	if (proxy && openconnect_set_http_proxy(vpninfo, strdup(proxy)))
 		exit(1);
 
-#ifndef _WIN32
+#if !defined(_WIN32) && !defined(__native_client__)
 	if (use_syslog) {
 		openlog("openconnect", LOG_PID, LOG_DAEMON);
 		vpninfo->progress = syslog_progress;
 	}
+#endif /* !_WIN32 && !__native_client__ */
 
+#ifndef _WIN32
 	memset(&sa, 0, sizeof(sa));
 
 	sa.sa_handler = handle_signal;
@@ -1418,31 +1499,8 @@ int main(int argc, char **argv)
 
 	if (!vpnc_script)
 		vpnc_script = xstrdup(default_vpncscript);
-#ifndef _WIN32
-	if (script_tun) {
-		if (openconnect_setup_tun_script(vpninfo, vpnc_script)) {
-			fprintf(stderr, _("Set up tun script failed\n"));
-			openconnect_vpninfo_free(vpninfo);
-			exit(1);
-		}
-	} else
-#endif
-	if (openconnect_setup_tun_device(vpninfo, vpnc_script, ifname)) {
-		fprintf(stderr, _("Set up tun device failed\n"));
-		openconnect_vpninfo_free(vpninfo);
-		exit(1);
-	}
 
-#ifndef _WIN32
-	if (uid != getuid()) {
-		if (setuid(uid)) {
-			fprintf(stderr, _("Failed to set uid %ld\n"),
-				(long)uid);
-			openconnect_vpninfo_free(vpninfo);
-			exit(1);
-		}
-	}
-#endif
+	STRDUP(vpninfo->vpnc_script, vpnc_script);
 
 	if (vpninfo->dtls_state != DTLS_DISABLED &&
 	    openconnect_setup_dtls(vpninfo, 60))
@@ -1466,7 +1524,7 @@ int main(int argc, char **argv)
 			compr = " + lz4";
 	}
 	vpn_progress(vpninfo, PRG_INFO,
-		     _("Connected %s as %s%s%s, using %s%s\n"), openconnect_get_ifname(vpninfo),
+		     _("Connected as %s%s%s, using %s%s\n"),
 		     ip_info->addr?:"",
 		     (ip_info->netmask6 && ip_info->addr) ? " + " : "",
 		     ip_info->netmask6 ? : "",
@@ -1660,7 +1718,7 @@ static int validate_peer_cert(void *_vpninfo, const char *reason)
 		fprintf(stderr, _("Enter '%s' to accept, '%s' to abort; anything else to view: "),
 		       _("yes"), _("no"));
 
-		read_stdin(&response, 0);
+		read_stdin(&response, 0, 0);
 		if (!response)
 			return -EINVAL;
 
@@ -1731,19 +1789,24 @@ static char *prompt_for_input(const char *prompt,
 			      struct openconnect_info *vpninfo,
 			      int hidden)
 {
-	char *response;
+	char *response = NULL;
 
 	fprintf(stderr, "%s", prompt);
 	fflush(stderr);
 
 	if (non_inter) {
-		fprintf(stderr, "***\n");
-		vpn_progress(vpninfo, PRG_ERR,
+		if (allow_stdin_read) {
+			read_stdin(&response, hidden, 1);
+		}
+		if (response == NULL) {
+			fprintf(stderr, "***\n");
+			vpn_progress(vpninfo, PRG_ERR,
 			     _("User input required in non-interactive mode\n"));
-		return NULL;
+		}
+		return response;
 	}
 
-	read_stdin(&response, hidden);
+	read_stdin(&response, hidden, 0);
 	return response;
 }
 
