@@ -22,6 +22,11 @@
 #include <stdlib.h>
 #include <unistd.h>
 #include <string.h>
+#ifndef _WIN32
+/* for setgroups() */
+# include <sys/types.h>
+# include <grp.h>
+#endif
 
 #include "openconnect-internal.h"
 
@@ -45,6 +50,12 @@ int tun_mainloop(struct openconnect_info *vpninfo, int *timeout)
 	struct pkt *this;
 	int work_done = 0;
 
+	if (!tun_is_up(vpninfo)) {
+		/* no tun yet, clear any queued packets */
+		while ((this = dequeue_packet(&vpninfo->incoming_queue)));
+		return 0;
+	}
+
 	if (read_fd_monitored(vpninfo, tun)) {
 		struct pkt *out_pkt = vpninfo->tun_pkt;
 		while (1) {
@@ -95,6 +106,60 @@ int tun_mainloop(struct openconnect_info *vpninfo, int *timeout)
 	return work_done;
 }
 
+static int setup_tun_device(struct openconnect_info *vpninfo)
+{
+	int ret;
+
+	if (vpninfo->setup_tun) {
+		vpninfo->setup_tun(vpninfo->cbdata);
+		if (tun_is_up(vpninfo))
+			return 0;
+	}
+
+#ifndef _WIN32
+	if (vpninfo->use_tun_script) {
+		ret = openconnect_setup_tun_script(vpninfo, vpninfo->vpnc_script);
+		if (ret) {
+			fprintf(stderr, _("Set up tun script failed\n"));
+			return ret;
+		}
+	} else
+#endif
+	ret = openconnect_setup_tun_device(vpninfo, vpninfo->vpnc_script, vpninfo->ifname);
+	if (ret) {
+		fprintf(stderr, _("Set up tun device failed\n"));
+		return ret;
+	}
+
+#if !defined(_WIN32) && !defined(__native_client__)
+	if (vpninfo->uid != getuid()) {
+		int e;
+
+		if (setgid(vpninfo->gid)) {
+			e = errno;
+			fprintf(stderr, _("Failed to set gid %ld: %s\n"),
+				(long)vpninfo->gid, strerror(e));
+			return -EPERM;
+		}
+
+		if (setgroups(1, &vpninfo->gid)) {
+			e = errno;
+			fprintf(stderr, _("Failed to set groups to %ld: %s\n"),
+				(long)vpninfo->gid, strerror(e));
+			return -EPERM;
+		}
+
+		if (setuid(vpninfo->uid)) {
+			e = errno;
+			fprintf(stderr, _("Failed to set uid %ld: %s\n"),
+				(long)vpninfo->uid, strerror(e));
+			return -EPERM;
+		}
+	}
+#endif
+	return 0;
+}
+
 /* Return value:
  *  = 0, when successfully paused (may call again)
  *  = -EINTR, if aborted locally via OC_CMD_CANCEL
@@ -119,7 +184,7 @@ int openconnect_mainloop(struct openconnect_info *vpninfo,
 
 	while (!vpninfo->quit_reason) {
 		int did_work = 0;
-		int timeout = INT_MAX;
+		int timeout;
 #ifdef _WIN32
 		HANDLE events[4];
 		int nr_events = 0;
@@ -128,14 +193,39 @@ int openconnect_mainloop(struct openconnect_info *vpninfo,
 		fd_set rfds, wfds, efds;
 #endif
 
+		/* If tun is not up, loop more often to detect
+		 * a DTLS timeout (due to a firewall block) as soon. */
+		if (tun_is_up(vpninfo))
+			timeout = INT_MAX;
+		else
+			timeout = 1000;
+
 		if (vpninfo->dtls_state > DTLS_DISABLED) {
-			ret = vpninfo->proto.udp_mainloop(vpninfo, &timeout);
+			/* Postpone tun device creation after DTLS is connected so
+			 * we have a better knowledge of the link MTU. We also
+			 * force the creation if DTLS enters sleeping mode - i.e.,
+			 * we failed to connect on time. */
+			if (!tun_is_up(vpninfo) && (vpninfo->dtls_state == DTLS_CONNECTED ||
+			    vpninfo->dtls_state == DTLS_SLEEPING)) {
+				ret = setup_tun_device(vpninfo);
+				if (ret) {
+					break;
+				}
+			}
+
+			ret = vpninfo->proto->udp_mainloop(vpninfo, &timeout);
 			if (vpninfo->quit_reason)
 				break;
 			did_work += ret;
+
+		} else if (!tun_is_up(vpninfo)) {
+			/* No DTLS - setup TUN device unconditionally */
+			ret = setup_tun_device(vpninfo);
+			if (ret)
+				break;
 		}
 
-		ret = vpninfo->proto.tcp_mainloop(vpninfo, &timeout);
+		ret = vpninfo->proto->tcp_mainloop(vpninfo, &timeout);
 		if (vpninfo->quit_reason)
 			break;
 		did_work += ret;
@@ -156,12 +246,13 @@ int openconnect_mainloop(struct openconnect_info *vpninfo,
 			}
 			break;
 		}
+
 		if (vpninfo->got_pause_cmd) {
 			/* close all connections and wait for the user to call
 			   openconnect_mainloop() again */
 			openconnect_close_https(vpninfo, 0);
-			if (vpninfo->dtls_state != DTLS_DISABLED) {
-				vpninfo->proto.udp_close(vpninfo);
+			if (vpninfo->dtls_state > DTLS_DISABLED) {
+				vpninfo->proto->udp_close(vpninfo);
 				vpninfo->dtls_state = DTLS_SLEEPING;
 				vpninfo->new_dtls_started = 0;
 			}
@@ -212,10 +303,11 @@ int openconnect_mainloop(struct openconnect_info *vpninfo,
 #endif
 	}
 
-	if (vpninfo->quit_reason && vpninfo->proto.vpn_close_session)
-		vpninfo->proto.vpn_close_session(vpninfo, vpninfo->quit_reason);
+	if (vpninfo->quit_reason && vpninfo->proto->vpn_close_session)
+		vpninfo->proto->vpn_close_session(vpninfo, vpninfo->quit_reason);
 
-	os_shutdown_tun(vpninfo);
+	if (tun_is_up(vpninfo))
+		os_shutdown_tun(vpninfo);
 	return ret < 0 ? ret : -EIO;
 }
 
