@@ -116,6 +116,7 @@
 #define SHA1_SIZE 20
 #define MD5_SIZE 16
 
+#define MAX(x,y) ((x)>(y))?(x):(y)
 /****************************************************************************/
 
 struct pkt {
@@ -151,9 +152,9 @@ struct pkt {
 #define KA_KEEPALIVE	3
 #define KA_REKEY	4
 
-#define DTLS_NOSECRET	0
-#define DTLS_SECRET	1
-#define DTLS_DISABLED	2
+#define DTLS_NOSECRET	0	/* Random secret has not been generated yet */
+#define DTLS_SECRET	1	/* Secret is present, ready to attempt DTLS */
+#define DTLS_DISABLED	2	/* DTLS was disabled on the *client* side */
 #define DTLS_SLEEPING	3	/* For ESP, sometimes sending probes */
 #define DTLS_CONNECTING	4	/* ESP probe received; must tell server */
 #define DTLS_CONNECTED	5	/* Server informed and should be sending ESP */
@@ -161,6 +162,7 @@ struct pkt {
 #define COMPR_DEFLATE	(1<<0)
 #define COMPR_LZS	(1<<1)
 #define COMPR_LZ4	(1<<2)
+#define COMPR_MAX	COMPR_LZ4
 
 #ifdef HAVE_LZ4
 #define COMPR_STATELESS	(COMPR_LZS | COMPR_LZ4)
@@ -244,6 +246,7 @@ struct http_auth_state {
 };
 
 struct vpn_proto {
+	const char *name;
 	int (*vpn_close_session)(struct openconnect_info *vpninfo, const char *reason);
 
 	/* This does the full authentication, calling back as appropriate */
@@ -315,8 +318,8 @@ struct esp {
 	gnutls_cipher_hd_t cipher;
 	gnutls_hmac_hd_t hmac;
 #elif defined(ESP_OPENSSL)
-	HMAC_CTX hmac;
-	EVP_CIPHER_CTX cipher;
+	HMAC_CTX *hmac, *pkt_hmac;
+	EVP_CIPHER_CTX *cipher;
 #endif
 	uint32_t seq;
 	uint32_t seq_backlog;
@@ -325,7 +328,7 @@ struct esp {
 };
 
 struct openconnect_info {
-	struct vpn_proto proto;
+	const struct vpn_proto *proto;
 
 #ifdef HAVE_ICONV
 	iconv_t ic_legacy_to_utf8;
@@ -398,9 +401,7 @@ struct openconnect_info {
 	int nopasswd;
 	int xmlpost;
 	char *dtls_ciphers;
-	uid_t uid_csd;
 	char *csd_wrapper;
-	int uid_csd_given;
 	int no_http_keepalive;
 	int dump_http_traffic;
 
@@ -444,6 +445,8 @@ struct openconnect_info {
 
 	void *peer_cert;
 	char *peer_cert_hash;
+	void *cert_list_handle;
+	int cert_list_size;
 
 	char *cookie; /* Pointer to within cookies list */
 	struct oc_vpn_option *cookies;
@@ -470,6 +473,7 @@ struct openconnect_info {
 	gnutls_session_t https_sess;
 	gnutls_certificate_credentials_t https_cred;
 	char local_cert_md5[MD5_SIZE * 2 + 1]; /* For CSD */
+	char gnutls_prio[256];
 #ifdef HAVE_TROUSERS
 	TSS_HCONTEXT tpm_context;
 	TSS_HKEY srk;
@@ -534,8 +538,17 @@ struct openconnect_info {
 
 	char *dtls_cipher;
 	char *vpnc_script;
+#ifndef _WIN32
+	int uid_csd_given;
+	uid_t uid_csd;
+	gid_t gid_csd;
+	uid_t uid;
+	gid_t gid;
+#endif
+	int use_tun_script;
 	int script_tun;
 	char *ifname;
+	char *cmd_ifname;
 
 	int reqmtu, basemtu;
 	const char *banner;
@@ -600,6 +613,9 @@ struct openconnect_info {
 	openconnect_process_auth_form_vfn process_auth_form;
 	openconnect_progress_vfn progress;
 	openconnect_protect_socket_vfn protect_socket;
+	openconnect_getaddrinfo_vfn getaddrinfo_override;
+	openconnect_setup_tun_vfn setup_tun;
+	openconnect_reconnected_vfn reconnected;
 
 	int (*ssl_read)(struct openconnect_info *vpninfo, char *buf, size_t len);
 	int (*ssl_gets)(struct openconnect_info *vpninfo, char *buf, size_t len);
@@ -704,6 +720,14 @@ static inline int set_fd_cloexec(int fd)
 	return fcntl(fd, F_SETFD, fcntl(fd, F_GETFD) | FD_CLOEXEC);
 #endif
 }
+static inline int tun_is_up(struct openconnect_info *vpninfo)
+{
+#ifdef _WIN32
+	return vpninfo->tun_fh != NULL;
+#else
+	return vpninfo->tun_fd != -1;
+#endif
+}
 
 #ifdef _WIN32
 #define pipe(fds) _pipe(fds, 4096, O_BINARY)
@@ -767,6 +791,7 @@ int dtls_setup(struct openconnect_info *vpninfo, int dtls_attempt_period);
 int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout);
 void dtls_close(struct openconnect_info *vpninfo);
 void dtls_shutdown(struct openconnect_info *vpninfo);
+void append_dtls_ciphers(struct openconnect_info *vpninfo, struct oc_text_buf *buf);
 
 /* cstp.c */
 void cstp_common_headers(struct openconnect_info *vpninfo, struct oc_text_buf *buf);
@@ -996,7 +1021,7 @@ static inline int strprefix_match(const char *str, int len, const char *match)
 }
 
 #define STRDUP(res, arg) \
-	do {							\
+	if (res != arg) {					\
 		free(res);					\
 		if (arg) {					\
 			res = strdup(arg);			\
