@@ -36,24 +36,40 @@
 #include <openssl/ui.h>
 #include <openssl/rsa.h>
 
+#if OPENSSL_VERSION_NUMBER < 0x10100000L
+#define X509_up_ref(x) 	CRYPTO_add(&(x)->references, 1, CRYPTO_LOCK_X509)
+#endif
+
+#if OPENSSL_VERSION_NUMBER < 0x10100000L
+#define EVP_MD_CTX_new EVP_MD_CTX_create
+#define EVP_MD_CTX_free EVP_MD_CTX_destroy
+#define X509_STORE_CTX_get0_chain(ctx) ((ctx)->chain)
+#define X509_STORE_CTX_get0_untrusted(ctx) ((ctx)->untrusted)
+#define X509_STORE_CTX_get0_cert(ctx) ((ctx)->cert)
+#endif
+
 int openconnect_sha1(unsigned char *result, void *data, int len)
 {
-	EVP_MD_CTX c;
+	EVP_MD_CTX *c = EVP_MD_CTX_new();
+
+	if (!c)
+		return -ENOMEM;
 
-	EVP_MD_CTX_init(&c);
 	EVP_Digest(data, len, result, NULL, EVP_sha1(), NULL);
-	EVP_MD_CTX_cleanup(&c);
+	EVP_MD_CTX_free(c);
 
 	return 0;
 }
 
 int openconnect_md5(unsigned char *result, void *data, int len)
 {
-	EVP_MD_CTX c;
+	EVP_MD_CTX *c = EVP_MD_CTX_new();
+
+	if (!c)
+		return -ENOMEM;
 
-	EVP_MD_CTX_init(&c);
 	EVP_Digest(data, len, result, NULL, EVP_md5(), NULL);
-	EVP_MD_CTX_cleanup(&c);
+	EVP_MD_CTX_free(c);
 
 	return 0;
 }
@@ -272,6 +288,7 @@ struct ui_form_opt {
 	UI_STRING *uis;
 };
 
+#ifdef HAVE_ENGINE
  /* Ick. But there is no way to pass this sanely through OpenSSL */
 static struct openconnect_info *ui_vpninfo;
 
@@ -414,6 +431,7 @@ static UI_METHOD *create_openssl_ui(struct openconnect_info *vpninfo)
 
 	return ui_method;
 }
+#endif
 
 static int pem_pw_cb(char *buf, int len, int w, void *v)
 {
@@ -464,7 +482,7 @@ static int install_extra_certs(struct openconnect_info *vpninfo, const char *sou
 					  buf, sizeof(buf));
 			vpn_progress(vpninfo, PRG_DEBUG,
 				     _("Extra cert from %s: '%s'\n"), source, buf);
-			CRYPTO_add(&cert2->references, 1, CRYPTO_LOCK_X509);
+			X509_up_ref(cert2);
 			SSL_CTX_add_extra_chain_cert(vpninfo->https_ctx, cert2);
 			cert = cert2;
 			goto next;
@@ -973,6 +991,7 @@ static int set_peer_cert_hash(struct openconnect_info *vpninfo)
 	return 0;
 }
 
+#if OPENSSL_VERSION_NUMBER < 0x10002000L
 static int match_hostname_elem(const char *hostname, int helem_len,
 			       const char *match, int melem_len)
 {
@@ -1041,32 +1060,16 @@ static int match_hostname(const char *hostname, const char *match)
 }
 
 /* cf. RFC2818 and RFC2459 */
-static int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert)
+static int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert,
+			       const unsigned char *ipaddr, int ipaddrlen)
 {
 	STACK_OF(GENERAL_NAME) *altnames;
 	X509_NAME *subjname;
 	ASN1_STRING *subjasn1;
 	char *subjstr = NULL;
-	int addrlen = 0;
 	int i, altdns = 0;
-	char addrbuf[sizeof(struct in6_addr)];
 	int ret;
 
-	/* Allow GEN_IP in the certificate only if we actually connected
-	   by IP address rather than by name. */
-	if (inet_pton(AF_INET, vpninfo->hostname, addrbuf) > 0)
-		addrlen = 4;
-	else if (inet_pton(AF_INET6, vpninfo->hostname, addrbuf) > 0)
-		addrlen = 16;
-	else if (vpninfo->hostname[0] == '[' &&
-		 vpninfo->hostname[strlen(vpninfo->hostname)-1] == ']') {
-		char *p = &vpninfo->hostname[strlen(vpninfo->hostname)-1];
-		*p = 0;
-		if (inet_pton(AF_INET6, vpninfo->hostname + 1, addrbuf) > 0)
-			addrlen = 16;
-		*p = ']';
-	}
-
 	altnames = X509_get_ext_d2i(peer_cert, NID_subject_alt_name,
 				    NULL, NULL);
 	for (i = 0; i < sk_GENERAL_NAME_num(altnames); i++) {
@@ -1098,7 +1101,7 @@ static int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert
 					     str);
 			}
 			OPENSSL_free(str);
-		} else if (this->type == GEN_IPADD && addrlen) {
+		} else if (this->type == GEN_IPADD && ipaddrlen) {
 			char host[80];
 			int family;
 
@@ -1116,8 +1119,8 @@ static int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert
 			/* We only do this for the debug messages */
 			inet_ntop(family, this->d.ip->data, host, sizeof(host));
 
-			if (this->d.ip->length == addrlen &&
-			    !memcmp(addrbuf, this->d.ip->data, addrlen)) {
+			if (this->d.ip->length == ipaddrlen &&
+			    !memcmp(ipaddr, this->d.ip->data, ipaddrlen)) {
 				vpn_progress(vpninfo, PRG_DEBUG,
 					     _("Matched %s address '%s'\n"),
 					     (family == AF_INET6) ? "IPv6" : "IPv4",
@@ -1156,7 +1159,7 @@ static int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert
 
 			/* Leave url_host as it was so that it can be freed */
 			url_host2 = url_host;
-			if (addrlen == 16 && vpninfo->hostname[0] != '[' &&
+			if (ipaddrlen == 16 && vpninfo->hostname[0] != '[' &&
 			    url_host[0] == '[' && url_host[strlen(url_host)-1] == ']') {
 				/* Cope with https://[IPv6]/ when the hostname is bare IPv6 */
 				url_host[strlen(url_host)-1] = 0;
@@ -1246,34 +1249,47 @@ static int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert
 	OPENSSL_free(subjstr);
 	return ret;
 }
-
-static int verify_peer(struct openconnect_info *vpninfo, SSL *https_ssl)
+#else
+static int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert,
+			       const unsigned char *ipaddr, int ipaddrlen)
 {
-	int ret;
-	int vfy = SSL_get_verify_result(https_ssl);
-	const char *err_string = NULL;
+	char *matched = NULL;
 
-	if (vfy != X509_V_OK)
-		err_string = X509_verify_cert_error_string(vfy);
-	else if (match_cert_hostname(vpninfo, vpninfo->peer_cert))
-		err_string = _("certificate does not match hostname");
+	if (ipaddrlen && X509_check_ip(peer_cert, ipaddr, ipaddrlen, 0)) {
+		if (vpninfo->verbose >= PRG_DEBUG) {
+			char host[80];
+			int family;
 
-	if (err_string) {
-		vpn_progress(vpninfo, PRG_INFO,
-			     _("Server certificate verify failed: %s\n"),
-			     err_string);
+			if (ipaddrlen == 4)
+				family = AF_INET;
+			else
+				family = AF_INET6;
 
-		if (vpninfo->validate_peer_cert)
-			ret = vpninfo->validate_peer_cert(vpninfo->cbdata,
-							  err_string);
-		else
-			ret = -EINVAL;
-	} else {
-		ret = 0;
+			inet_ntop(family, ipaddr, host, sizeof(host));
+			vpn_progress(vpninfo, PRG_DEBUG,
+				     _("Matched %s address '%s'\n"),
+				     (family == AF_INET6) ? "IPv6" : "IPv4",
+				     host);
+		}
+		return 0;
+	}
+	if (X509_check_host(peer_cert, vpninfo->hostname, 0, 0, &matched)) {
+		vpn_progress(vpninfo, PRG_DEBUG,
+			     _("Matched peer certificate subject name '%s'\n"),
+			     matched);
+		OPENSSL_free(matched);
+		return 0;
 	}
 
-	return ret;
+	/* We do it like this because these two strings are already
+	 * translated in gnutls.c */
+	vpn_progress(vpninfo, PRG_INFO,
+		     _("Server certificate verify failed: %s\n"),
+		     _("certificate does not match hostname"));
+
+	return -EINVAL;
 }
+#endif /* OpenSSL < 1.0.2 */
 
 /* Before OpenSSL 1.1 we could do this directly. And needed to. */
 #ifndef SSL_CTX_get_extra_chain_certs_only
@@ -1284,15 +1300,18 @@ static int verify_peer(struct openconnect_info *vpninfo, SSL *https_ssl)
 static void workaround_openssl_certchain_bug(struct openconnect_info *vpninfo,
 					     SSL *ssl)
 {
-	/* OpenSSL has problems with certificate chains -- if there are
+	/* OpenSSL has problems with certificate chains — if there are
 	   multiple certs with the same name, it doesn't necessarily
-	   choose the _right_ one. (RT#1942)
-	   Pick the right ones for ourselves and add them manually. */
+	   choose the _right_ one. (RT#1942). This affects some servers,
+	   and we can work around it by picking the right ones ourselves
+	   and adding them manually on the wire. */
 	X509 *cert = SSL_get_certificate(ssl);
 	X509 *cert2;
 	X509_STORE *store = SSL_CTX_get_cert_store(vpninfo->https_ctx);
-	X509_STORE_CTX ctx;
+	X509_STORE_CTX *ctx;
 	void *extra_certs;
+	STACK_OF(X509) *chain;
+	int i;
 
 	if (!cert || !store)
 		return;
@@ -1302,11 +1321,22 @@ static void workaround_openssl_certchain_bug(struct openconnect_info *vpninfo,
 	if (extra_certs)
 		return;
 
-	if (!X509_STORE_CTX_init(&ctx, store, NULL, NULL))
+	ctx = X509_STORE_CTX_new();
+	if (!ctx)
 		return;
+	if (X509_STORE_CTX_init(ctx, store, NULL, NULL))
+		goto out;
 
-	while (ctx.get_issuer(&cert2, &ctx, cert) == 1) {
+	X509_STORE_CTX_set_cert(ctx, cert);
+
+	/* We only really want it to build the chain; we don't really
+	   care if it's trusted */
+	(void)X509_verify_cert(ctx);
+
+	chain = X509_STORE_CTX_get0_chain(ctx);
+	for (i = 1; i < sk_X509_num(chain); i++) {
 		char buf[200];
+		cert2 = sk_X509_value(chain, i);
 		if (cert2 == cert)
 			break;
 		if (X509_check_issued(cert2, cert2) == X509_V_OK)
@@ -1316,20 +1346,120 @@ static void workaround_openssl_certchain_bug(struct openconnect_info *vpninfo,
 				  buf, sizeof(buf));
 		vpn_progress(vpninfo, PRG_DEBUG,
 			     _("Extra cert from cafile: '%s'\n"), buf);
+		X509_up_ref(cert);
 		SSL_CTX_add_extra_chain_cert(vpninfo->https_ctx, cert);
 	}
-	X509_STORE_CTX_cleanup(&ctx);
+out:
+	X509_STORE_CTX_free(ctx);
+}
+
+int openconnect_get_peer_cert_chain(struct openconnect_info *vpninfo,
+				    struct oc_cert **chainp)
+{
+	struct oc_cert *chain, *p;
+	X509_STORE_CTX *ctx = vpninfo->cert_list_handle;
+	STACK_OF(X509) *untrusted = X509_STORE_CTX_get0_untrusted(ctx);
+	int i, cert_list_size;
+
+	if (!ctx)
+		return -EINVAL;
+
+	cert_list_size = sk_X509_num(untrusted);
+	if (!cert_list_size)
+		return -EIO;
+
+	p = chain = calloc(cert_list_size, sizeof(struct oc_cert));
+	if (!chain)
+		return -ENOMEM;
+
+	for (i = 0; i < cert_list_size; i++, p++) {
+		X509 *cert = sk_X509_value(untrusted, i);
+
+		p->der_len = i2d_X509(cert, &p->der_data);
+		if (p->der_len < 0) {
+			openconnect_free_peer_cert_chain(vpninfo, chain);
+			return -ENOMEM;
+		}
+	}
+
+	*chainp = chain;
+	return cert_list_size;
+}
+
+void openconnect_free_peer_cert_chain(struct openconnect_info *vpninfo,
+				      struct oc_cert *chain)
+{
+	int i;
+
+	for (i = 0; i < vpninfo->cert_list_size; i++)
+		OPENSSL_free(chain[i].der_data);
+	free(chain);
 }
 
-#if OPENSSL_VERSION_NUMBER >= 0x00908000
 static int ssl_app_verify_callback(X509_STORE_CTX *ctx, void *arg)
 {
-	/* We've seen certificates in the wild which don't have the
-	   purpose fields filled in correctly */
-	X509_VERIFY_PARAM_set_purpose(ctx->param, X509_PURPOSE_ANY);
-	return X509_verify_cert(ctx);
+	struct openconnect_info *vpninfo = arg;
+	const char *err_string = NULL;
+	X509 *cert = X509_STORE_CTX_get0_cert(ctx);
+
+	if (vpninfo->peer_cert) {
+		/* This is a *rehandshake*. Require that the server
+		 * presents exactly the same certificate as the
+		 * first time. */
+		if (X509_cmp(cert, vpninfo->peer_cert)) {
+			vpn_progress(vpninfo, PRG_ERR, _("Server presented different cert on rehandshake\n"));
+			return 0;
+		}
+		vpn_progress(vpninfo, PRG_TRACE, _("Server presented identical cert on rehandshake\n"));
+		return 1;
+	}
+	vpninfo->peer_cert = cert;
+	X509_up_ref(cert);
+
+	set_peer_cert_hash(vpninfo);
+
+	if (!X509_verify_cert(ctx)) {
+		err_string = X509_verify_cert_error_string(X509_STORE_CTX_get_error(ctx));
+	} else {
+		unsigned char addrbuf[sizeof(struct in6_addr)];
+		int addrlen = 0;
+
+		if (inet_pton(AF_INET, vpninfo->hostname, addrbuf) > 0)
+			addrlen = 4;
+		else if (inet_pton(AF_INET6, vpninfo->hostname, addrbuf) > 0)
+			addrlen = 16;
+		else if (vpninfo->hostname[0] == '[' &&
+			 vpninfo->hostname[strlen(vpninfo->hostname)-1] == ']') {
+			char *p = &vpninfo->hostname[strlen(vpninfo->hostname)-1];
+			*p = 0;
+			if (inet_pton(AF_INET6, vpninfo->hostname + 1, addrbuf) > 0)
+				addrlen = 16;
+			*p = ']';
+		}
+
+		if (match_cert_hostname(vpninfo, vpninfo->peer_cert, addrbuf, addrlen))
+			err_string = _("certificate does not match hostname");
+		else
+			return 1;
+	}
+
+	vpn_progress(vpninfo, PRG_INFO,
+		     _("Server certificate verify failed: %s\n"),
+		     err_string);
+
+	if (vpninfo->validate_peer_cert) {
+		int ret;
+
+		vpninfo->cert_list_handle = ctx;
+		ret = vpninfo->validate_peer_cert(vpninfo->cbdata, err_string);
+		vpninfo->cert_list_handle = NULL;
+
+		if (!ret)
+			return 1;
+	}
+
+	return 0;
 }
-#endif
 
 static int check_certificate_expiry(struct openconnect_info *vpninfo)
 {
@@ -1421,18 +1551,12 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 			check_certificate_expiry(vpninfo);
 		}
 
-		/* We just want to do:
-		   SSL_CTX_set_purpose(vpninfo->https_ctx, X509_PURPOSE_ANY);
-		   ... but it doesn't work with OpenSSL < 0.9.8k because of
-		   problems with inheritance (fixed in v1.1.4.6 of
-		   crypto/x509/x509_vpm.c) so we have to play silly buggers
-		   instead. This trick doesn't work _either_ in < 0.9.7 but
-		   I don't know of _any_ workaround which will, and can't
-		   be bothered to find out either. */
-#if OPENSSL_VERSION_NUMBER >= 0x00908000
+		/* We've seen certificates in the wild which don't have the
+		   purpose fields filled in correctly */
+		SSL_CTX_set_purpose(vpninfo->https_ctx, X509_PURPOSE_ANY);
 		SSL_CTX_set_cert_verify_callback(vpninfo->https_ctx,
-						 ssl_app_verify_callback, NULL);
-#endif
+						 ssl_app_verify_callback, vpninfo);
+
 		if (!vpninfo->no_system_trust)
 			SSL_CTX_set_default_verify_paths(vpninfo->https_ctx);
 
@@ -1508,14 +1632,32 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 	https_bio = BIO_new_socket(ssl_sock, BIO_NOCLOSE);
 	BIO_set_nbio(https_bio, 1);
 	SSL_set_bio(https_ssl, https_bio, https_bio);
-
-#if 0
-	/* Should be enabled on openssl versions that support
-	   SSL_set_tlsext_host_name() after the F5 firewall workaround
-	   is enabled */
+	/*
+	 * If a ClientHello is between 256 and 511 bytes, the
+	 * server cannot distinguish between a SSLv2 formatted
+	 * packet and a SSLv3 formatted packet.
+	 *
+	 * F5 BIG-IP reverse proxies in particular will
+	 * silently drop an ambiguous ClientHello.
+	 *
+	 * OpenSSL fixes this in v1.0.1g+ by padding ClientHello
+	 * packets to at least 512 bytes.
+	 *
+	 * For older versions of OpenSSL, we try to avoid long
+	 * packets by silently disabling extensions such as SNI.
+	 *
+	 * Discussion:
+	 * http://www.ietf.org/mail-archive/web/tls/current/msg10423.html
+	 *
+	 * OpenSSL commits:
+	 * 4fcdd66fff5fea0cfa1055c6680a76a4303f28a2
+	 * cd6bd5ffda616822b52104fee0c4c7d623fd4f53
+	 */
+#if OPENSSL_VERSION_NUMBER >= 0x10001070
 	if (string_is_hostname(vpninfo->hostname))
 		SSL_set_tlsext_host_name(https_ssl, vpninfo->hostname);
 #endif
+	SSL_set_verify(https_ssl, SSL_VERIFY_PEER, NULL);
 
 	vpn_progress(vpninfo, PRG_INFO, _("SSL negotiation with %s\n"),
 		     vpninfo->hostname);
@@ -1551,14 +1693,6 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 	}
 
 	vpninfo->cstp_cipher = (char *)SSL_get_cipher_name(https_ssl);
-	vpninfo->peer_cert = SSL_get_peer_certificate(https_ssl);
-	set_peer_cert_hash(vpninfo);
-
-	if (verify_peer(vpninfo, https_ssl)) {
-		SSL_free(https_ssl);
-		closesocket(ssl_sock);
-		return -EINVAL;
-	}
 
 	vpninfo->ssl_fd = ssl_sock;
 	vpninfo->https_ssl = https_ssl;
