@@ -57,38 +57,78 @@ static inline int connect_pending()
 	return errno == EINPROGRESS;
 #endif
 }
+
+/* Windows is interminably horrid, and has disjoint errno spaces.
+ * So if we return a positive value, that's a WSA Error and should
+ * be handled with openconnect__win32_strerror(). But if we return a
+ * negative value, that's a normal errno and should be handled with
+ * strerror(). No, you can't just pass the latter value (negated) to
+ * openconnect__win32_strerror() because it gives nonsense results. */
 static int cancellable_connect(struct openconnect_info *vpninfo, int sockfd,
 			       const struct sockaddr *addr, socklen_t addrlen)
 {
 	struct sockaddr_storage peer;
 	socklen_t peerlen = sizeof(peer);
-	fd_set wr_set, rd_set;
+	fd_set wr_set, rd_set, ex_set;
 	int maxfd = sockfd;
+	int err;
 
 	set_sock_nonblock(sockfd);
 	if (vpninfo->protect_socket)
 		vpninfo->protect_socket(vpninfo->cbdata, sockfd);
 
-	if (connect(sockfd, addr, addrlen) < 0 && !connect_pending())
-		return -1;
+	if (connect(sockfd, addr, addrlen) < 0 && !connect_pending()) {
+#ifdef _WIN32
+		return WSAGetLastError();
+#else
+		return -errno;
+#endif
+	}
 
 	do {
 		FD_ZERO(&wr_set);
 		FD_ZERO(&rd_set);
+		FD_ZERO(&ex_set);
 		FD_SET(sockfd, &wr_set);
+#ifdef _WIN32 /* Windows indicates failure this way, not in wr_set */
+		FD_SET(sockfd, &ex_set);
+#endif
 		cmd_fd_set(vpninfo, &rd_set, &maxfd);
-
-		select(maxfd + 1, &rd_set, &wr_set, NULL, NULL);
+		select(maxfd + 1, &rd_set, &wr_set, &ex_set, NULL);
 		if (is_cancel_pending(vpninfo, &rd_set)) {
 			vpn_progress(vpninfo, PRG_ERR, _("Socket connect cancelled\n"));
-			errno = EINTR;
-			return -1;
+			return -EINTR;
 		}
-	} while (!FD_ISSET(sockfd, &wr_set) && !vpninfo->got_pause_cmd);
+	} while (!FD_ISSET(sockfd, &wr_set) && !FD_ISSET(sockfd, &ex_set) &&
+		 !vpninfo->got_pause_cmd);
 
 	/* Check whether connect() succeeded or failed by using
 	   getpeername(). See http://cr.yp.to/docs/connect.html */
-	return getpeername(sockfd, (void *)&peer, &peerlen);
+	if (!getpeername(sockfd, (void *)&peer, &peerlen))
+		return 0;
+
+#ifdef _WIN32 /* On Windows, use getsockopt() to determine the error.
+	       * We don't ddo this on Windows because it just reports
+	       * -ENOTCONN, which we already knew. */
+
+	err = WSAGetLastError();
+	if (err == WSAENOTCONN) {
+		socklen_t errlen = sizeof(err);
+
+		getsockopt(sockfd, SOL_SOCKET, SO_ERROR,
+			   (void *)&err, &errlen);
+	}
+#else
+	err = -errno;
+	if (err == -ENOTCONN) {
+		int ch;
+
+		if (read(sockfd, &ch, 1) < 0)
+			err = -errno;
+		/* It should *always* fail! */
+	}
+#endif
+	return err;
 }
 
 /* checks whether the provided string is an IP or a hostname.
@@ -143,21 +183,39 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 #endif
 		{
 			ssl_sock = socket(vpninfo->peer_addr->sa_family, SOCK_STREAM, IPPROTO_IP);
-			if (ssl_sock < 0)
+			if (ssl_sock < 0) {
+#ifdef _WIN32
+				err = WSAGetLastError();
+#else
+				err = -errno;
+#endif
 				goto reconn_err;
+			}
 			set_fd_cloexec(ssl_sock);
 		}
-		if (cancellable_connect(vpninfo, ssl_sock, vpninfo->peer_addr, vpninfo->peer_addrlen)) {
+		err = cancellable_connect(vpninfo, ssl_sock, vpninfo->peer_addr, vpninfo->peer_addrlen);
+		if (err) {
+			char *errstr;
 		reconn_err:
+#ifdef _WIN32
+			if (err > 0)
+				errstr = openconnect__win32_strerror(err);
+			else
+#endif
+				errstr = strerror(-err);
 			if (vpninfo->proxy) {
 				vpn_progress(vpninfo, PRG_ERR,
-					     _("Failed to reconnect to proxy %s\n"),
-					     vpninfo->proxy);
+					     _("Failed to reconnect to proxy %s: %s\n"),
+					     vpninfo->proxy, errstr);
 			} else {
 				vpn_progress(vpninfo, PRG_ERR,
-					     _("Failed to reconnect to host %s\n"),
-					     vpninfo->hostname);
+					     _("Failed to reconnect to host %s: %s\n"),
+					     vpninfo->hostname, errstr);
 			}
+#ifdef _WIN32
+			if (err > 0)
+				free(errstr);
+#endif
 			if (ssl_sock >= 0)
 				closesocket(ssl_sock);
 			ssl_sock = -EINVAL;
@@ -242,7 +300,10 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 			hints.ai_flags |= AI_NUMERICHOST;
 		}
 
-		err = getaddrinfo(hostname, port, &hints, &result);
+		if (vpninfo->getaddrinfo_override)
+			err = vpninfo->getaddrinfo_override(vpninfo->cbdata, hostname, port, &hints, &result);
+		else
+			err = getaddrinfo(hostname, port, &hints, &result);
 
 		if (err) {
 			vpn_progress(vpninfo, PRG_ERR,
@@ -269,7 +330,7 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 			host[0] = 0;
 			if (!getnameinfo(rp->ai_addr, rp->ai_addrlen, host,
 					 sizeof(host), NULL, 0, NI_NUMERICHOST))
-				vpn_progress(vpninfo, PRG_INFO, vpninfo->proxy_type ?
+				vpn_progress(vpninfo, PRG_DEBUG, vpninfo->proxy_type ?
 						     _("Attempting to connect to proxy %s%s%s:%s\n") :
 						     _("Attempting to connect to server %s%s%s:%s\n"),
 					     rp->ai_family == AF_INET6 ? "[" : "",
@@ -282,9 +343,22 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 			if (ssl_sock < 0)
 				continue;
 			set_fd_cloexec(ssl_sock);
-			if (cancellable_connect(vpninfo, ssl_sock, rp->ai_addr, rp->ai_addrlen) >= 0) {
+			err = cancellable_connect(vpninfo, ssl_sock, rp->ai_addr, rp->ai_addrlen);
+			if (!err) {
 				/* Store the peer address we actually used, so that DTLS can
 				   use it again later */
+				free(vpninfo->ip_info.gateway_addr);
+				vpninfo->ip_info.gateway_addr = NULL;
+
+				if (host[0]) {
+					vpninfo->ip_info.gateway_addr = strdup(host);
+					vpn_progress(vpninfo, PRG_INFO, _("Connected to %s%s%s:%s\n"),
+						     rp->ai_family == AF_INET6 ? "[" : "",
+						     host,
+						     rp->ai_family == AF_INET6 ? "]" : "",
+						     port);
+				}
+
 				free(vpninfo->peer_addr);
 				vpninfo->peer_addrlen = 0;
 				vpninfo->peer_addr = malloc(rp->ai_addrlen);
@@ -323,6 +397,25 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 				}
 				break;
 			}
+			if (host[0]) {
+				char *errstr;
+#ifdef _WIN32
+				if (err > 0)
+					errstr = openconnect__win32_strerror(err);
+				else
+#endif
+					errstr = strerror(-err);
+
+				vpn_progress(vpninfo, PRG_INFO, _("Failed to connect to %s%s%s:%s: %s\n"),
+					     rp->ai_family == AF_INET6 ? "[" : "",
+					     host,
+					     rp->ai_family == AF_INET6 ? "]" : "",
+					     port, errstr);
+#ifdef _WIN32
+				if (err > 0)
+					free(errstr);
+#endif
+			}
 			closesocket(ssl_sock);
 			ssl_sock = -1;
 
@@ -335,6 +428,8 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 				free(vpninfo->peer_addr);
 				vpninfo->peer_addr = 0;
 				vpninfo->peer_addrlen = 0;
+				free(vpninfo->ip_info.gateway_addr);
+				vpninfo->ip_info.gateway_addr = NULL;
 			}
 		}
 		freeaddrinfo(result);
@@ -487,7 +582,7 @@ int openconnect_passphrase_from_fsid(struct openconnect_info *vpninfo)
 
 	return 0;
 }
-#else
+#elif defined(HAVE_STATFS)
 int openconnect_passphrase_from_fsid(struct openconnect_info *vpninfo)
 {
 	char *sslkey = openconnect_utf8_to_legacy(vpninfo, vpninfo->sslkey);
@@ -513,6 +608,11 @@ int openconnect_passphrase_from_fsid(struct openconnect_info *vpninfo)
 
 	return err;
 }
+#else
+int openconnect_passphrase_from_fsid(struct openconnect_info *vpninfo)
+{
+	return -EOPNOTSUPP;
+}
 #endif
 
 #if defined(OPENCONNECT_OPENSSL) || defined(DTLS_OPENSSL)
@@ -706,7 +806,7 @@ void check_cmd_fd(struct openconnect_info *vpninfo, fd_set *fds)
 int is_cancel_pending(struct openconnect_info *vpninfo, fd_set *fds)
 {
 	check_cmd_fd(vpninfo, fds);
-	return vpninfo->got_cancel_cmd;
+	return vpninfo->got_cancel_cmd || vpninfo->got_pause_cmd;
 }
 
 void poll_cmd_fd(struct openconnect_info *vpninfo, int timeout)
@@ -903,7 +1003,7 @@ int ssl_reconnect(struct openconnect_info *vpninfo)
 	free(vpninfo->tun_pkt);
 	vpninfo->tun_pkt = NULL;
 
-	while ((ret = vpninfo->proto.tcp_connect(vpninfo))) {
+	while ((ret = vpninfo->proto->tcp_connect(vpninfo))) {
 		if (timeout <= 0)
 			return ret;
 		if (ret == -EPERM) {
@@ -924,6 +1024,10 @@ int ssl_reconnect(struct openconnect_info *vpninfo)
 		if (interval > RECONNECT_INTERVAL_MAX)
 			interval = RECONNECT_INTERVAL_MAX;
 	}
+
 	script_config_tun(vpninfo, "reconnect");
+	if (vpninfo->reconnected)
+		vpninfo->reconnected(vpninfo->cbdata);
+
 	return 0;
 }
